"""
Copyright 2017-2019 Government of Canada - Public Services and Procurement Canada - buyandsell.gc.ca

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


import asyncio
import json
import logging

from typing import Union

from indy import did, ledger
from indy.error import IndyError, ErrorCode

from von_anchor.cache import RevoCacheEntry, CRED_DEF_CACHE, ENDPOINT_CACHE, REVO_CACHE, SCHEMA_CACHE
from von_anchor.error import (
    AbsentCredDef,
    AbsentPool,
    AbsentRecord,
    AbsentRevReg,
    AbsentNym,
    AbsentSchema,
    BadIdentifier,
    BadLedgerTxn,
    ClosedPool,
    CorruptWallet,
    WalletState)
from von_anchor.indytween import Role, SchemaKey
from von_anchor.nodepool import NodePool
from von_anchor.util import ok_cred_def_id, ok_did, ok_rev_reg_id, ok_schema_id, schema_id, schema_key
from von_anchor.wallet import EndpointInfo, Wallet


LOGGER = logging.getLogger(__name__)


class BaseAnchor:
    """
    Base class for common anchor functionality. A VON anchor has a wallet and a
    node pool.  It has a cryptonym and can interact via indy-sdk
    with the distributed ledger that its node pool operates.
    """

    def __init__(self, wallet: Wallet, pool: NodePool = None, **kwargs) -> None:
        """
        Initializer for anchor. Retain wallet and node pool.

        :param wallet: wallet for anchor use
        :param pool: optional node pool for anchor use
        :param kwargs: place holders for super(); implementation ignores
        """

        LOGGER.debug('BaseAnchor.__init__ >>> wallet: %s, pool: %s, kwargs: %s', wallet, pool, kwargs)

        self._wallet = wallet
        self._pool = pool

        LOGGER.debug('BaseAnchor.__init__ <<<')

    @property
    def pool(self) -> NodePool:
        """
        Accessor for node pool.

        :return: node pool
        """

        return self._pool

    @pool.setter
    def pool(self, value: NodePool) -> None:
        """
        Set pool.

        :param value: pool
        """

        self._pool = value

    @property
    def wallet(self) -> Wallet:
        """
        Accessor for wallet.

        :return: wallet
        """

        return self._wallet

    @wallet.setter
    def wallet(self, value: Wallet) -> None:
        """
        Set wallet.

        :param value: wallet
        """

        self._wallet = value

    @property
    def name(self) -> str:
        """
        Accessor for anchor name, from wallet.

        :return: anchor (wallet) name
        """

        return self.wallet.name

    @property
    def did(self) -> str:
        """
        Accessor for anchor DID.

        :return: anchor DID
        """

        return self.wallet.did

    @property
    def verkey(self) -> str:
        """
        Accessor for anchor verification key.

        :return: anchor verification key
        """

        return self.wallet.verkey

    async def __aenter__(self) -> 'BaseAnchor':
        """
        Context manager entry.
        For use in monolithic call opening, using, and closing the anchor.

        :return: current object
        """

        LOGGER.debug('BaseAnchor.__aenter__ >>>')

        rv = await self.open()

        LOGGER.debug('BaseAnchor.__aenter__ <<<')
        return rv

    async def open(self) -> 'BaseAnchor':
        """
        Context manager entry.
        For use when keeping anchor open across multiple calls.

        :return: current object
        """

        LOGGER.debug('BaseAnchor.open >>>')

        # Do not open pool independently: let relying party decide when to go on-line and off-line
        # Do not open wallet independently: allow for sharing open wallet over many anchor lifetimes

        LOGGER.debug('BaseAnchor.open <<<')
        return self

    async def __aexit__(self, exc_type, exc, traceback) -> None:
        """
        Context manager exit.
        For use in monolithic call opening, using, and closing the anchor.

        :param exc_type:
        :param exc:
        :param traceback:
        """

        LOGGER.debug('BaseAnchor.__aexit__ >>> exc_type: %s, exc: %s, traceback: %s', exc_type, exc, traceback)

        await self.close()

        LOGGER.debug('BaseAnchor.__exit__ <<<')

    async def close(self) -> None:
        """
        Explicit exit.
        For use when keeping anchor open across multiple calls.
        """

        LOGGER.debug('BaseAnchor.close >>>')

        # Do not close wallet independently: allow for sharing open wallet over many anchor lifetimes
        # Do not close pool independently: let relying party decide when to go on-line and off-line

        LOGGER.debug('BaseAnchor.close <<<')

    async def reseed(self, seed: str = None) -> None:
        """
        Rotate key for VON anchor: generate new key, submit to ledger, update wallet.
        Raise WalletState if wallet is currently closed.

        :param seed: new seed for ed25519 key pair (default random)
        """

        LOGGER.debug('BaseAnchor.reseed >>> seed: [SEED]')

        verkey = await self.wallet.reseed_init(seed)
        req_json = await ledger.build_nym_request(
            self.did,
            self.did,
            verkey,
            self.name,
            (await self.get_nym_role()).token())
        await self._sign_submit(req_json)
        await self.wallet.reseed_apply()

        LOGGER.debug('BaseAnchor.reseed <<<')

    async def get_nym(self, target_did: str = None) -> str:
        """
        Get json cryptonym (including current verification key) for input (anchor) DID from ledger.
        Return empty production {} if the ledger has no such cryptonym.

        Raise BadLedgerTxn on failure. Raise WalletState if target DID is default (own DID) value but
        wallet does not have it (neither created nor opened since initialization).

        :param target_did: DID of cryptonym to fetch (default own DID)
        :return: cryptonym json
        """

        LOGGER.debug('BaseAnchor.get_nym >>> target_did: %s', target_did)

        if target_did and not ok_did(target_did):
            LOGGER.debug('BaseAnchor.get_nym <!< Bad DID %s', target_did)
            raise BadIdentifier('Bad DID {}'.format(target_did))

        if not (target_did or self.did):
            LOGGER.debug('BaseAnchor.get_nym <!< Bad wallet state: DID for %s unavailable', self.name)
            raise WalletState('Bad wallet state: DID for {} unavailable'.format(self.name))

        rv = json.dumps({})
        get_nym_req = await ledger.build_get_nym_request(self.did, target_did or self.did)
        resp_json = await self._submit(get_nym_req)

        data_json = (json.loads(resp_json))['result']['data']  # it's double-encoded on the ledger
        if data_json:
            rv = data_json

        LOGGER.debug('BaseAnchor.get_nym <<< %s', rv)
        return rv

    async def get_nym_role(self, target_did: str = None) -> Role:
        """
        Return the cryptonym role for input did from the ledger - note that this may exceed
        the role of least privilege for the class.

        Raise AbsentNym if current anchor has no cryptonym on the ledger, or WalletState if current DID unavailable.

        :param target_did: DID of cryptonym role to fetch (default own DID)
        :return: identifier for current cryptonym role on ledger
        """

        LOGGER.debug('BaseAnchor.get_nym_role >>> target_did: %s', target_did)

        nym = json.loads(await self.get_nym(target_did))
        if not nym:
            LOGGER.debug('BaseAnchor.get_nym_role <!< Ledger has no cryptonym for anchor %s', self.name)
            raise AbsentNym('Ledger has no cryptonym for anchor {}'.format(self.name))

        rv = Role.get(nym['role'])

        LOGGER.debug('BaseAnchor.get_nym_role <<< %s', rv)
        return rv

    @staticmethod
    def least_role() -> Role:
        """
        Return the indy-sdk role of least privilege for an anchor (class) in building
        its cryptonym for the trust anchor to send to the ledger.

        :return: role of least privilege by anchor class
        """

        LOGGER.debug('BaseAnchor.least_role >>>')

        rv = Role.TRUST_ANCHOR

        LOGGER.debug('BaseAnchor.least_role <<< %s', rv)
        return rv

    async def set_did_endpoint(self, remote_did: str, did_endpoint: str) ->  EndpointInfo:
        """
        Set endpoint as metadata for pairwise remote DID in wallet. Pick up (transport)
        verification key from pairwise relation and return with endpoint in EndpointInfo.


        Raise BadIdentifier on bad DID. Raise WalletState if wallet is closed.
        Raise AbsentRecord if pairwise relation not present in wallet.

        :param remote_did: pairwise remote DID
        :param endpoint: value to set as endpoint in wallet and cache
        :return: endpoint and (transport) verification key
        """

        LOGGER.debug('BaseAnchor.set_did_endpoint >>> remote_did: %s, did_endpoint: %s', remote_did, did_endpoint)

        if not ok_did(remote_did):
            LOGGER.debug('BaseAnchor.set_did_endpoint <!< Bad DID %s', remote_did)
            raise BadIdentifier('Bad DID {}'.format(remote_did))

        pairwise_info = (await self.wallet.get_pairwise(remote_did)).get(remote_did, None)
        if not pairwise_info:
            LOGGER.debug(
                'BaseAnchor.set_did_endpoint <!< Anchor %s has no pairwise relation for remote DID %s',
                self.name,
                remote_did)
            raise AbsentRecord('Anchor {} has no pairwise relation for remote DID {}'.format(
                self.name,
                remote_did))

        await self.wallet.write_pairwise(
            pairwise_info.their_did,
            pairwise_info.their_verkey,
            pairwise_info.my_did,
            {'did_endpoint': did_endpoint})

        rv = EndpointInfo(did_endpoint, pairwise_info.their_verkey)

        LOGGER.debug('BaseAnchor.set_did_endpoint <<< %s', rv)
        return rv

    async def get_did_endpoint(self, remote_did: str) -> EndpointInfo:
        """
        Return endpoint info for remote DID.
        Raise BadIdentifier for bad remote DID. Raise WalletState if bypassing cache but wallet is closed.
        Raise AbsentRecord for no such endpoint.

        :param remote_did: pairwise remote DID
        :return: endpoint and (transport) verification key as EndpointInfo
        """

        LOGGER.debug('BaseAnchor.get_did_endpoint >>> remote_did: %s', remote_did)

        if not ok_did(remote_did):
            LOGGER.debug('BaseAnchor.get_did_endpoint <!< Bad DID %s', remote_did)
            raise BadIdentifier('Bad DID {}'.format(remote_did))

        pairwise_info = (await self.wallet.get_pairwise(remote_did)).get(remote_did, None)
        if not (pairwise_info and 'did_endpoint' in pairwise_info.metadata):
            LOGGER.debug('BaseAnchor.get_did_endpoint <!< No endpoint for remote DID %s', remote_did)
            raise AbsentRecord('No endpoint for remote DID {}'.format(remote_did))
        rv = EndpointInfo(pairwise_info.metadata['did_endpoint'], pairwise_info.their_verkey)

        LOGGER.debug('BaseAnchor.get_did_endpoint <<< %s', rv)
        return rv

    async def send_endpoint(self, endpoint: str) -> None:
        """
        Send anchor's own endpoint attribute to ledger (and endpoint cache),
        if ledger does not yet have input value. Specify None to clear.

        Raise BadIdentifier on endpoint not formatted as '<ip-address>:<port>',
        BadLedgerTxn on failure, WalletState if wallet is closed.

        :param endpoint: value to set as endpoint attribute on ledger and cache:
            specify URL or None to clear.
        """

        LOGGER.debug('BaseAnchor.send_endpoint >>> endpoint: %s', endpoint)

        ledger_endpoint = await self.get_endpoint()
        if ledger_endpoint == endpoint:
            LOGGER.info('%s endpoint already set as %s', self.name, endpoint)
            LOGGER.debug('BaseAnchor.send_endpoint <<< (%s already set for %s )')
            return

        attr_json = json.dumps({
            'endpoint': {
                'endpoint': endpoint  # indy-sdk likes 'ha' here but won't map 'ha' to a URL, only ip:port
            }
        })
        req_json = await ledger.build_attrib_request(self.did, self.did, None, attr_json, None)
        await self._sign_submit(req_json)

        for _ in range(16):  # reasonable timeout
            if await self.get_endpoint(None, False) == endpoint:
                break
            await asyncio.sleep(1)
            LOGGER.info('Sent endpoint %s to ledger, waiting 1s for its confirmation', endpoint)
        else:
            LOGGER.debug('BaseAnchor.send_endpoint <!< timed out waiting on send endpoint %s', endpoint)
            raise BadLedgerTxn('Timed out waiting on sent endpoint {}'.format(endpoint))

        LOGGER.debug('BaseAnchor.send_endpoint <<<')

    async def get_endpoint(self, target_did: str = None, from_cache: bool = True) -> str:
        """
        Get endpoint attribute for anchor having input DID (default own DID).
        Raise WalletState if target DID is default (own DID) value but wallet does not have it
        (neither created nor opened since initialization).

        :param target_did: DID of anchor for which to find endpoint attribute on ledger
        :param from_cache: check endpoint cache first before visiting ledger; always update cache with ledger value
        :return: endpoint attribute value, or None for no such value
        """

        LOGGER.debug('BaseAnchor.get_endpoint >>> target_did: %s, from_cache: %s', target_did, from_cache)

        rv = None

        if not (target_did or self.did):
            LOGGER.debug('BaseAnchor.get_endpoint <!< Bad wallet state: DID for %s unavailable', self.name)
            raise WalletState('Bad wallet state: DID for {} unavailable'.format(self.name))

        target_did = target_did or self.did
        if not ok_did(target_did):
            LOGGER.debug('BaseAnchor.get_endpoint <!< Bad DID %s', target_did)
            raise BadIdentifier('Bad DID {}'.format(target_did))

        if from_cache:
            with ENDPOINT_CACHE.lock:
                if target_did in ENDPOINT_CACHE:
                    LOGGER.info('BaseAnchor.get_endpoint: got endpoint for %s from cache', target_did)
                    rv = ENDPOINT_CACHE[target_did]
                    LOGGER.debug('BaseAnchor.get_endpoint <<< %s', rv)
                    return rv

        req_json = await ledger.build_get_attrib_request(
            self.did,
            target_did,
            'endpoint',
            None,
            None)
        resp_json = await self._submit(req_json)

        data_json = (json.loads(resp_json))['result']['data']  # it's double-encoded on the ledger
        if data_json:
            rv = json.loads(data_json)['endpoint'].get('endpoint', None)
        else:
            LOGGER.info('_AgentCore.get_endpoint: ledger query returned response with no data')

        with ENDPOINT_CACHE.lock:
            if rv:
                ENDPOINT_CACHE[target_did] = rv
            else:
                ENDPOINT_CACHE.pop(target_did, None)
                assert target_did not in ENDPOINT_CACHE

        LOGGER.debug('BaseAnchor.get_endpoint <<< %s', rv)
        return rv

    async def _submit(self, req_json: str) -> str:
        """
        Submit (json) request to ledger; return (json) result.

        Raise AbsentPool for no pool, ClosedPool if pool is not yet open, or BadLedgerTxn on failure.

        :param req_json: json of request to sign and submit
        :return: json response
        """

        LOGGER.debug('BaseAnchor._submit >>> req_json: %s', req_json)

        if not self.pool:
            LOGGER.debug('BaseAnchor._submit <!< absent pool')
            raise AbsentPool('Cannot submit request: absent pool')

        if not self.pool.handle:
            LOGGER.debug('BaseAnchor._submit <!< closed pool %s', self.pool.name)
            raise ClosedPool('Cannot submit request to closed pool {}'.format(self.pool.name))

        rv_json = await ledger.submit_request(self.pool.handle, req_json)
        await asyncio.sleep(0)

        resp = json.loads(rv_json)
        if resp.get('op', '') in ('REQNACK', 'REJECT'):
            LOGGER.debug('BaseAnchor._submit <!< ledger rejected request: %s', resp['reason'])
            raise BadLedgerTxn('Ledger rejected transaction request: {}'.format(resp['reason']))

        LOGGER.debug('BaseAnchor._submit <<< %s', rv_json)
        return rv_json

    async def _sign_submit(self, req_json: str) -> str:
        """
        Sign and submit (json) request to ledger; return (json) result.

        Raise:
        * AbsentPool for no pool
        * ClosedPool if pool is not yet open
        * CorruptWallet if existing wallet appears not to pertain to the anchor's pool
        * WalletState if wallet is closed
        * BadLedgerTxn on ledger rejection of transaction.

        :param req_json: json of request to sign and submit
        :return: json response
        """

        LOGGER.debug('BaseAnchor._sign_submit >>> req_json: %s', req_json)

        if not self.pool:
            LOGGER.debug('BaseAnchor._sign_submit <!< absent pool')
            raise AbsentPool('Cannot sign and submit request: absent pool')

        if not self.pool.handle:
            LOGGER.debug('BaseAnchor._submit <!< closed pool %s', self.pool.name)
            raise ClosedPool('Cannot sign and submit request to closed pool {}'.format(self.pool.name))

        if not self.wallet.handle:
            LOGGER.debug('BaseAnchor._sign_submit <!< Wallet %s is closed', self.name)
            raise WalletState('Wallet {} is closed'.format(self.name))

        try:
            rv_json = await ledger.sign_and_submit_request(self.pool.handle, self.wallet.handle, self.did, req_json)
            await asyncio.sleep(0)
        except IndyError as x_indy:
            if x_indy.error_code == ErrorCode.WalletIncompatiblePoolError:
                LOGGER.debug(
                    'BaseAnchor._sign_submit <!< Corrupt wallet %s is not compatible with pool %s',
                    self.name,
                    self.pool.name)
                raise CorruptWallet('Corrupt wallet {} is not compatible with pool {}'.format(
                    self.name,
                    self.pool.name))
            LOGGER.debug(
                'BaseAnchor._sign_submit <!< cannot sign/submit request for ledger: indy error code %s',
                x_indy.error_code)
            raise BadLedgerTxn('Cannot sign/submit request for ledger: indy error code {}'.format(
                x_indy.error_code))

        resp = json.loads(rv_json)
        if resp.get('op', '') in ('REQNACK', 'REJECT'):
            LOGGER.debug('BaseAnchor._sign_submit: ledger rejected request: %s', resp['reason'])
            raise BadLedgerTxn('Ledger rejected transaction request: {}'.format(resp['reason']))

        LOGGER.debug('BaseAnchor._sign_submit <<< %s', rv_json)
        return rv_json

    async def _verkey_for(self, target: str) -> str:
        """
        Given a DID, retrieve its verification key, looking in wallet, then pool.
        Given a verification key or None, return input.

        Raise WalletState if the wallet is closed. Given a recipient DID not in the wallet,
        raise AbsentPool if the instance has no pool or ClosedPool if its pool is closed.
        If no such verification key is on the ledger, raise AbsentNym.

        :param target: verification key, or DID to resolve to such
        :return: verification key
        """

        LOGGER.debug('BaseAnchor._verkey_for >>> target: %s', target)

        rv = target
        if rv is None or not ok_did(rv):  # it's None or already a verification key
            LOGGER.debug('BaseAnchor._verkey_for <<< %s', rv)
            return rv

        if self.wallet.handle:
            try:
                rv = await did.key_for_local_did(self.wallet.handle, target)
                LOGGER.info('Anchor %s got verkey for DID %s from wallet', self.name, target)
                LOGGER.debug('BaseAnchor._verkey_for <<< %s', rv)
                return rv
            except IndyError as x_indy:
                if x_indy.error_code != ErrorCode.WalletItemNotFound:  # on not found, try the pool
                    LOGGER.debug(
                        'BaseAnchor._verkey_for <!< key lookup for local DID %s raised indy error code %s',
                        target,
                        x_indy.error_code)
                    raise

        nym = json.loads(await self.get_nym(target))
        if not nym:
            LOGGER.debug(
                'BaseAnchor._verkey_for <!< Anchor %s cannot get cryptonym (hence verkey) for DID %s',
                self.name,
                target)
            raise AbsentNym('Anchor {} cannot get cryptonym (hence verkey) for DID {}'.format(self.name, target))

        rv = json.loads(await self.get_nym(target))['verkey']
        LOGGER.info('Anchor %s got verkey for DID %s from pool %s', self.name, target, self.pool.name)

        LOGGER.debug('BaseAnchor._verkey_for <<< %s', rv)
        return rv

    async def get_rev_reg_def(self, rr_id: str) -> str:
        """
        Get revocation registry definition from ledger by its identifier. Raise AbsentRevReg
        for no such revocation registry, logging any error condition and raising BadLedgerTxn
        on bad request.

        Retrieve the revocation registry definition from the anchor's revocation cache if it has it;
        cache it en passant if it does not (and such a revocation registry definition exists on the ledger).

        :param rr_id: (revocation registry) identifier string, of the format
            '<issuer-did>:4:<issuer-did>:3:CL:<schema-seq-no>:<tag>:CL_ACCUM:<tag>'
        :return: revocation registry definition json as retrieved from ledger
        """

        LOGGER.debug('BaseAnchor.get_rev_reg_def >>> rr_id: %s', rr_id)

        if not ok_rev_reg_id(rr_id):
            LOGGER.debug('BaseAnchor.get_rev_reg_def <!< Bad rev reg id %s', rr_id)
            raise BadIdentifier('Bad rev reg id {}'.format(rr_id))

        rv_json = json.dumps({})

        with REVO_CACHE.lock:
            revo_cache_entry = REVO_CACHE.get(rr_id, None)
            rr_def = revo_cache_entry.rev_reg_def if revo_cache_entry else None
            if rr_def:
                LOGGER.info('BaseAnchor.get_rev_reg_def: rev reg def for %s from cache', rr_id)
                rv_json = json.dumps(rr_def)
            else:
                get_rr_def_req_json = await ledger.build_get_revoc_reg_def_request(self.did, rr_id)
                resp_json = await self._submit(get_rr_def_req_json)
                try:
                    (_, rv_json) = await ledger.parse_get_revoc_reg_def_response(resp_json)
                    rr_def = json.loads(rv_json)
                except IndyError:  # ledger replied, but there is no such rev reg
                    LOGGER.debug('BaseAnchor.get_rev_reg_def <!< no rev reg exists on %s', rr_id)
                    raise AbsentRevReg('No rev reg exists on {}'.format(rr_id))

                if revo_cache_entry is None:
                    REVO_CACHE[rr_id] = RevoCacheEntry(rr_def, None)
                else:
                    REVO_CACHE[rr_id].rev_reg_def = rr_def

        LOGGER.debug('BaseAnchor.get_rev_reg_def <<< %s', rv_json)
        return rv_json

    async def get_cred_def(self, cd_id: str) -> str:
        """
        Get credential definition from ledger by its identifier.

        Raise AbsentCredDef for no such credential definition, logging any error condition and raising
        BadLedgerTxn on bad request. Raise ClosedPool if cred def not in cache and pool is closed.

        Retrieve the credential definition from the anchor's credential definition cache if it has it; cache it
        en passant if it does not (and if there is a corresponding credential definition on the ledger).

        :param cd_id: (credential definition) identifier string ('<issuer-did>:3:CL:<schema-seq-no>:<tag>')
        :return: credential definition json as retrieved from ledger, empty production for no such cred def
        """

        LOGGER.debug('BaseAnchor.get_cred_def >>> cd_id: %s', cd_id)

        if not ok_cred_def_id(cd_id):
            LOGGER.debug('BaseAnchor.get_cred_def <!< Bad cred def id %s', cd_id)
            raise BadIdentifier('Bad cred def id {}'.format(cd_id))

        rv_json = json.dumps({})

        with CRED_DEF_CACHE.lock:
            if cd_id in CRED_DEF_CACHE:
                LOGGER.info('BaseAnchor.get_cred_def: got cred def for %s from cache', cd_id)
                rv_json = json.dumps(CRED_DEF_CACHE[cd_id])
                LOGGER.debug('BaseAnchor.get_cred_def <<< %s', rv_json)
                return rv_json

            req_json = await ledger.build_get_cred_def_request(self.did, cd_id)
            resp_json = await self._submit(req_json)
            resp = json.loads(resp_json)
            if not ('result' in resp and resp['result'].get('data', None)):
                LOGGER.debug('BaseAnchor.get_cred_def <!< no cred def exists on %s', cd_id)
                raise AbsentCredDef('No cred def exists on {}'.format(cd_id))
            try:
                (_, rv_json) = await ledger.parse_get_cred_def_response(resp_json)
            except IndyError:  # ledger replied, but there is no such cred def
                LOGGER.debug('BaseAnchor.get_cred_def <!< no cred def exists on %s', cd_id)
                raise AbsentCredDef('No cred def exists on {}'.format(cd_id))
            CRED_DEF_CACHE[cd_id] = json.loads(rv_json)
            LOGGER.info('BaseAnchor.get_cred_def: got cred def %s from ledger', cd_id)

        LOGGER.debug('BaseAnchor.get_cred_def <<< %s', rv_json)
        return rv_json

    async def get_schema(self, index: Union[SchemaKey, int, str]) -> str:
        """
        Get schema from ledger by SchemaKey namedtuple (origin DID, name, version),
        sequence number, or schema identifier.

        Raise AbsentSchema for no such schema, logging any error condition and raising
        BadLedgerTxn on bad request.

        Retrieve the schema from the anchor's schema cache if it has it; cache it
        en passant if it does not (and there is a corresponding schema on the ledger).

        :param index: schema key (origin DID, name, version), sequence number, or schema identifier
        :return: schema json, parsed from ledger
        """

        LOGGER.debug('BaseAnchor.get_schema >>> index: %s', index)

        rv_json = json.dumps({})
        with SCHEMA_CACHE.lock:
            if SCHEMA_CACHE.contains(index):
                LOGGER.info('BaseAnchor.get_schema: got schema %s from cache', index)
                rv_json = SCHEMA_CACHE[index]
                LOGGER.debug('BaseAnchor.get_schema <<< %s', rv_json)
                return json.dumps(rv_json)

            if isinstance(index, SchemaKey) or (isinstance(index, str) and ok_schema_id(index)):
                s_id = schema_id(*index) if isinstance(index, SchemaKey) else index
                s_key = schema_key(s_id)
                req_json = await ledger.build_get_schema_request(self.did, s_id)
                resp_json = await self._submit(req_json)
                resp = json.loads(resp_json)

                if not ('result' in resp and resp['result'].get('data', {}).get('attr_names', None)):
                    LOGGER.debug('BaseAnchor.get_schema <!< no schema exists on %s', index)
                    raise AbsentSchema('No schema exists on {}'.format(index))
                try:
                    (_, rv_json) = await ledger.parse_get_schema_response(resp_json)
                except IndyError:  # ledger replied, but there is no such schema
                    LOGGER.debug('BaseAnchor.get_schema <!< no schema exists on %s', index)
                    raise AbsentSchema('No schema exists on {}'.format(index))
                SCHEMA_CACHE[s_key] = json.loads(rv_json)  # cache indexes by both txn# and schema key en passant
                LOGGER.info('BaseAnchor.get_schema: got schema %s from ledger', index)

            elif isinstance(index, (int, str)):  # index is not a schema id: it's a stringified int txn# if it's a str
                txn_json = await self.get_txn(int(index))  # raises AbsentPool if anchor has no pool
                txn = json.loads(txn_json)
                if txn.get('type', None) == '101':  # {} for no such txn; 101 marks indy-sdk schema txn type
                    rv_json = await self.get_schema(self.pool.protocol.txn_data2schema_key(txn))
                else:
                    LOGGER.debug('BaseAnchor.get_schema <!< no schema at seq #%s on ledger', index)
                    raise AbsentSchema('No schema at seq #{} on ledger'.format(index))

            else:
                LOGGER.debug('BaseAnchor.get_schema <!< bad schema index type')
                raise AbsentSchema('Attempt to get schema on ({}) {} , must use schema key or an int'.format(
                    type(index),
                    index))

        LOGGER.debug('BaseAnchor.get_schema <<< %s', rv_json)
        return rv_json

    async def encrypt(self, message: bytes, authn: bool = False, recip: str = None) -> bytes:
        """
        Encrypt plaintext for owner of DID or verification key, anonymously or via
        authenticated encryption scheme. If given DID, first check wallet and then pool
        for corresponding verification key.

        Raise WalletState if the wallet is closed. Given a recipient DID not in the wallet,
        raise AbsentPool if the instance has no pool or ClosedPool if its pool is closed.

        :param message: plaintext, as bytes
        :param authn: whether to use authenticated encryption scheme
        :param recip: DID or verification key of recipient, None for anchor's own
        :return: ciphertext, as bytes
        """

        LOGGER.debug('BaseAnchor.encrypt >>> message: %s, authn: %s, recip: %s', message, authn, recip)

        if not self.wallet.handle:  # don't risk fetching verkey from ledger only to fail encryption on closed wallet
            LOGGER.debug('BaseAnchor.encrypt <!< Wallet %s is closed', self.name)
            raise WalletState('Wallet {} is closed'.format(self.name))

        rv = await self.wallet.encrypt(message, authn, await self._verkey_for(recip))

        LOGGER.debug('BaseAnchor.auth_encrypt <<< %s', rv)
        return rv

    async def decrypt(self, ciphertext: bytes, sender: str = None) -> (bytes, str):
        """
        Decrypt ciphertext and optionally authenticate sender.

        Raise BadKey if authentication operation reveals sender key distinct from current
        verification key of owner of input DID.  Raise WalletState if wallet is closed.

        :param ciphertext: ciphertext, as bytes
        :param sender: DID or verification key of sender, None for anonymously encrypted ciphertext
        :return: decrypted bytes and sender verification key (None for anonymous decryption)
        """

        LOGGER.debug('BaseAnchor.decrypt >>> ciphertext: %s, sender: %s', ciphertext, sender)

        if not self.wallet.handle:  # don't risk fetching verkey from ledger only to fail encryption on closed wallet
            LOGGER.debug('BaseAnchor.decrypt <!< Wallet %s is closed', self.name)
            raise WalletState('Wallet {} is closed'.format(self.name))

        from_verkey = None
        if sender:
            from_verkey = await self._verkey_for(sender)
        rv = await self.wallet.decrypt(
            ciphertext,
            True if from_verkey else None,
            to_verkey=None,
            from_verkey=from_verkey)

        LOGGER.debug('BaseAnchor.decrypt <<< %s', rv)
        return rv

    async def sign(self, message: bytes) -> bytes:
        """
        Sign message; return signature. Raise WalletState if wallet is closed.

        :param message: Content to sign, as bytes
        :return: signature, as bytes
        """

        LOGGER.debug('BaseAnchor.sign >>> message: %s', message)

        if not self.wallet.handle:
            LOGGER.debug('BaseAnchor.sign <!< Wallet %s is closed', self.name)
            raise WalletState('Wallet {} is closed'.format(self.name))

        rv = await self.wallet.sign(message)

        LOGGER.debug('BaseAnchor.sign <<< %s', rv)
        return rv

    async def verify(self, message: bytes, signature: bytes, signer: str = None) -> bool:
        """
        Verify signature with input signer verification key (via lookup by DID first if need be).
        Raise WalletState if wallet is closed.

        :param message: Content to sign, as bytes
        :param signature: signature, as bytes
        :param signer: signer DID or verification key; omit for anchor's own
        :return: whether signature is valid
        """

        LOGGER.debug('BaseAnchor.verify >>> signer: %s, message: %s, signature: %s', signer, message, signature)

        if not self.wallet.handle:
            LOGGER.debug('BaseAnchor.verify <!< Wallet %s is closed', self.name)
            raise WalletState('Wallet {} is closed'.format(self.name))

        verkey = None
        if signer:
            verkey = await self._verkey_for(signer)
        rv = await self.wallet.verify(message, signature, verkey)

        LOGGER.debug('BaseAnchor.verify <<< %s', rv)
        return rv

    async def get_txn(self, seq_no: int) -> str:
        """
        Find a transaction on the distributed ledger by its sequence number.

        :param seq_no: transaction number
        :return: json sequence number of transaction, null for no match
        """

        LOGGER.debug('BaseAnchor.get_txn >>> seq_no: %s', seq_no)

        rv_json = json.dumps({})
        req_json = await ledger.build_get_txn_request(self.did, None, seq_no)
        resp = json.loads(await self._submit(req_json))

        rv_json = self.pool.protocol.txn2data(resp)

        LOGGER.debug('BaseAnchor.get_txn <<< %s', rv_json)
        return rv_json

    def __str__(self) -> str:
        """
        Return string representation for current object.

        :return: string representation for current object
        """

        return '{}({})'.format(self.__class__.__name__, self.name)



"""
Copyright 2017-2019 Government of Canada - Public Services and Procurement Canada - buyandsell.gc.ca

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


import logging

from collections import namedtuple
from os import sys
from os.path import dirname, realpath
from sys import exit as sys_exit, path as sys_path
from typing import Sequence

DIR_VON_ANCHOR = dirname(dirname(dirname(realpath(__file__))))
if DIR_VON_ANCHOR not in sys_path:
    sys_path.append(DIR_VON_ANCHOR)

from von_anchor import NominalAnchor, TrusteeAnchor
from von_anchor.error import AbsentNym, AbsentPool, BadRole, ExtantWallet, VonAnchorError
from von_anchor.frill import do_wait, inis2dict
from von_anchor.indytween import Role
from von_anchor.nodepool import NodePoolManager
from von_anchor.util import ok_role
from von_anchor.wallet import WalletManager


NodePoolData = namedtuple('NodePoolData', 'name genesis_txn_path')
AnchorData = namedtuple('AnchorData', 'role name seed did wallet_create wallet_type wallet_access')


def usage() -> None:
    """
    Print usage advice.
    """

    print()
    print('Usage: setnym.py <config-ini>')
    print()
    print('where <config-ini> represents the path to the configuration file.')
    print()
    print('The operation submits a nym to a trustee anchor to send to the ledger,')
    print('if the ledger does not have it already as configured.')
    print()
    print('The configuration file has sections and entries as follows:')
    print('  * section [Node Pool]:')
    print('    - name: the name of the node pool to which the operation applies')
    print('    - genesis.txn.path: the path to the genesis transaction file')
    print('        for the node pool (may omit if node pool already exists)')
    print('  * section [Trustee Anchor]:')
    print("    - name: the trustee anchor's (wallet) name")
    print("    - wallet.type: (default blank) the trustee anchor's wallet type")
    print("    - wallet.access: (default blank) the trustee anchor's")
    print('        wallet access credential (password) value')
    print('  * section [VON Anchor]:')
    print('    - role: the role to request in the send-nym transaction; specify:')
    print('        - (default) empty value for user with no additional write privileges')
    print('        - TRUST_ANCHOR for VON anchor with write privileges for indy artifacts')
    print('        - TRUSTEE for VON anchor sending further cryptonyms to the ledger')
    print("    - name: the VON anchor's (wallet) name")
    print("    - seed: the VON anchor's seed (optional, for wallet creation only)")
    print("    - did: the VON anchor's DID (optional, for wallet creation only)")
    print('    - wallet.create: whether create the wallet if it does not yet exist')
    print('        (value True/False, 1/0, or Yes/No)')
    print("    - wallet.type: (default blank) the VON anchor's wallet type")
    print("    - wallet.access: (default blank) the VON anchor's")
    print('        wallet access credential (password) value.')
    print()


async def _set_wallets(an_data: dict) -> dict:
    """
    Set wallets as configured for setnym operation.

    :param an_data: dict mapping profiles to anchor data
    :return: dict mapping anchor names to wallet objects
    """

    w_mgr = WalletManager()
    rv = {}
    for profile in an_data:
        w_cfg = {'id': an_data[profile].name}
        if an_data[profile].wallet_type:
            w_cfg['storage_type'] = an_data[profile].wallet_type
        if an_data[profile].seed:
            w_cfg['seed'] = an_data[profile].seed
        if an_data[profile].did:
            w_cfg['did'] = an_data[profile].did
        if an_data[profile].wallet_create:
            try:
                await w_mgr.create(w_cfg, access=an_data[profile].wallet_access)
            except ExtantWallet:
                pass
        rv[profile] = w_mgr.get(w_cfg, access=an_data[profile].wallet_access)

    return rv


async def setnym(ini_path: str) -> int:
    """
    Set configuration. Open pool, trustee anchor, and wallet of anchor whose nym to send.
    Register exit hooks to close pool and trustee anchor.

    Engage trustee anchor to send nym for VON anchor, if it differs on the ledger from configuration.

    :param ini_path: path to configuration file
    :return: 0 for OK, 1 for failure
    """

    config = inis2dict(ini_path)
    if config['Trustee Anchor']['name'] == config['VON Anchor']['name']:
        raise ExtantWallet('Wallet names must differ between VON Anchor and Trustee Anchor')

    cfg_van_role = config['VON Anchor'].get('role', None) or None  # nudge empty value from '' to None
    if not ok_role(cfg_van_role):
        raise BadRole('Configured role {} is not valid'.format(cfg_van_role))

    pool_data = NodePoolData(
        config['Node Pool']['name'],
        config['Node Pool'].get('genesis.txn.path', None) or None)

    an_data = {
        'tan': AnchorData(
            Role.TRUSTEE,
            config['Trustee Anchor']['name'],
            config['Trustee Anchor'].get('seed', None) or None,
            config['Trustee Anchor'].get('did', None) or None,
            config['Trustee Anchor'].get('wallet.create', '0').lower() in ['1', 'true', 'yes'],
            config['Trustee Anchor'].get('wallet.type', None) or None,
            config['Trustee Anchor'].get('wallet.access', None) or None),
        'van': AnchorData(
            Role.get(cfg_van_role),
            config['VON Anchor']['name'],
            config['VON Anchor'].get('seed', None) or None,
            config['VON Anchor'].get('did', None) or None,
            config['VON Anchor'].get('wallet.create', '0').lower() in ['1', 'true', 'yes'],
            config['VON Anchor'].get('wallet.type', None) or None,
            config['VON Anchor'].get('wallet.access', None) or None)
    }

    an_wallet = await _set_wallets(an_data)

    p_mgr = NodePoolManager()
    if pool_data.name not in await p_mgr.list():
        if pool_data.genesis_txn_path:
            await p_mgr.add_config(pool_data.name, pool_data.genesis_txn_path)
        else:
            raise AbsentPool('Node pool {} has no ledger configuration, but {} specifies no genesis txn path'.format(
                pool_data.name,
                ini_path))

    async with an_wallet['tan'] as w_tan, (
            an_wallet['van']) as w_van, (
            p_mgr.get(pool_data.name)) as pool, (
            TrusteeAnchor(w_tan, pool)) as tan, (
            NominalAnchor(w_van, pool)) as van:

        send_verkey = van.verkey
        try:
            nym_role = await tan.get_nym_role(van.did)
            if an_data['van'].role == nym_role:
                return 0  # ledger is as per configuration
            send_verkey = None  # only owner can touch verkey
            if nym_role != Role.USER:  # only remove role when it is not already None on the ledger
                await tan.send_nym(van.did, send_verkey, van.wallet.name, Role.ROLE_REMOVE)
        except AbsentNym:
            pass  # cryptonym not there yet, fall through

        await tan.send_nym(van.did, send_verkey, van.wallet.name, an_data['van'].role)

    return 0


def main(args: Sequence[str] = None) -> int:
    """
    Main line for script: check arguments and dispatch operation to set nym.

    :param args: command-line arguments
    :return: 0 for OK, 1 for failure
    """

    logging.basicConfig(
        level=logging.INFO,
        format='%(asctime)-15s | %(levelname)-8s | %(message)s',
        datefmt='%Y-%m-%d %H:%M:%S')
    logging.getLogger('von_anchor').setLevel(logging.WARNING)
    logging.getLogger('indy').setLevel(logging.ERROR)

    if args is None:
        args = sys.argv[1:]

    if len(sys.argv) == 2:
        try:
            return do_wait(setnym(sys.argv[1]))
        except VonAnchorError as vax:
            print(str(vax))
            return 1
    else:
        usage()
        return 1

if __name__ == '__main__':
    sys_exit(main())



"""
Copyright 2017-2019 Government of Canada - Public Services and Procurement Canada - buyandsell.gc.ca

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


import json
import logging

from indy import pool
from indy.error import IndyError, ErrorCode

from von_anchor.error import AbsentPool
from von_anchor.nodepool.protocol import Protocol
from von_anchor.validcfg import validate_config


LOGGER = logging.getLogger(__name__)


class NodePool:
    """
    Class encapsulating indy-sdk node pool.
    """

    def __init__(self, name: str, protocol: Protocol = None, config: dict = None) -> None:
        """
        Initializer for node pool. Does not open the pool, only retains input parameters.

        :param name: name of the pool
        :param protocol: indy-node protocol
        :param config: configuration, None for default
        """

        LOGGER.debug('NodePool.__init__ >>> name: %s, protocol: %s, config: %s', name, protocol, config)

        self._protocol = protocol or Protocol.DEFAULT
        self._name = name
        self._handle = None
        self._config = config or {}
        validate_config('pool', self._config)

        LOGGER.debug('NodePool.__init__ <<<')

    @property
    def name(self) -> str:
        """
        Accessor for pool name.

        :return: pool name
        """

        return self._name

    @property
    def handle(self) -> int:
        """
        Accessor for indy-sdk pool handle.

        :return: indy-sdk pool handle
        """

        return self._handle

    @property
    def config(self) -> dict:
        """
        Accessor for pool config.

        :return: pool config
        """

        return self._config

    @property
    def protocol(self) -> str:
        """
        Accessor for protocol version pool config setting.

        :return: protocol version pool config setting
        """

        return self._protocol

    async def __aenter__(self) -> 'NodePool':
        """
        Context manager entry. Opens pool as configured, for closure on context manager exit.
        Creates pool if it does not yet exist, using configured genesis transaction file.
        For use in monolithic call opening, using, and closing the pool.

        :return: current object
        """

        LOGGER.debug('NodePool.__aenter__ >>>')

        rv = await self.open()

        LOGGER.debug('NodePool.__aenter__ <<<')
        return rv

    async def open(self) -> 'NodePool':
        """
        Explicit entry. Opens pool as configured, for later closure via close().
        Creates pool if it does not yet exist, using configured genesis transaction file.
        For use when keeping pool open across multiple calls.

        Raise any AbsentPool if node pool ledger configuration is not available.

        :return: current object
        """

        LOGGER.debug('NodePool.open >>>')

        await pool.set_protocol_version(self.protocol.indy())
        LOGGER.info('Pool ledger %s set protocol %s', self.name, self.protocol)

        try:
            self._handle = await pool.open_pool_ledger(self.name, json.dumps(self.config))
        except IndyError as x_indy:
            if x_indy.error_code == ErrorCode.PoolLedgerNotCreatedError:
                LOGGER.debug('NodePool.open <!< Absent node pool %s ledger configuration', self.name)
                raise AbsentPool('Absent node pool {} ledger configuration'.format(self.name))
            LOGGER.debug(
                'NodePool.open <!< cannot open node pool %s: indy error code %s',
                self.name,
                x_indy.error_code)
            raise

        LOGGER.debug('NodePool.open <<<')
        return self

    async def __aexit__(self, exc_type, exc, traceback) -> None:
        """
        Context manager exit. Closes pool. For use in monolithic call opening, using, and closing the pool.

        Raise any IndyError causing failure to create ledger configuration.

        :param exc_type:
        :param exc:
        :param traceback:
        """

        LOGGER.debug('NodePool.__aexit__ >>>')

        await self.close()

        LOGGER.debug('NodePool.__aexit__ <<<')

    async def close(self) -> None:
        """
        Explicit exit. Closes pool. For use when keeping pool open across multiple calls.
        """

        LOGGER.debug('NodePool.close >>>')

        if not self.handle:
            LOGGER.warning('Abstaining from closing pool %s: already closed', self.name)
        else:
            await pool.close_pool_ledger(self.handle)
        self._handle = None

        LOGGER.debug('NodePool.close <<<')

    async def refresh(self) -> None:
        """
        Refresh local copy of pool ledger and update node pool connections.
        """

        LOGGER.debug('NodePool.refresh >>>')

        await pool.refresh_pool_ledger(self.handle)

        LOGGER.debug('NodePool.refresh <<<')


    def __repr__(self) -> str:
        """
        Return representation for current object.

        :return: representation for current object
        """

        return 'NodePool({}, {}, {})'.format(self.name, self.protocol, self.config)



"""
Copyright 2017-2019 Government of Canada - Public Services and Procurement Canada - buyandsell.gc.ca

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


import logging

from indy import ledger

from von_anchor.anchor.base import BaseAnchor
from von_anchor.error import BadIdentifier
from von_anchor.indytween import Role
from von_anchor.util import ok_did


LOGGER = logging.getLogger(__name__)


class AnchorSmith(BaseAnchor):
    """
    Mixin for (trustee) anchor to write anchors onto the distributed ledger.
    """

    @staticmethod
    def least_role() -> Role:
        """
        Return the TRUSTEE indy-sdk role for an anchor acting in an AnchorSmith capacity.

        :return: TRUSTEE role
        """

        LOGGER.debug('AnchorSmith.least_role >>>')

        rv = Role.TRUSTEE

        LOGGER.debug('AnchorSmith.least_role <<< %s', rv)
        return rv

    async def send_nym(self, did: str, verkey: str = None, alias: str = None, role: Role = None) -> None:
        """
        Send input anchor's cryptonym (including DID, verification key, plus optional alias and role)
        to the distributed ledger.

        Raise BadLedgerTxn on failure, BadIdentifier for bad DID, or BadRole for bad role.

        :param did: anchor DID to send to ledger
        :param verkey: optional anchor verification key
        :param alias: optional alias
        :param role: anchor role on the ledger (default value of USER)
        """

        LOGGER.debug(
            'AnchorSmith.send_nym >>> did: %s, verkey: %s, alias: %s, role: %s', did, verkey, alias, role)

        if not ok_did(did):
            LOGGER.debug('AnchorSmith.send_nym <!< Bad DID %s', did)
            raise BadIdentifier('Bad DID {}'.format(did))

        req_json = await ledger.build_nym_request(self.did, did, verkey, alias, (role or Role.USER).token())
        await self._sign_submit(req_json)

        LOGGER.debug('AnchorSmith.send_nym <<<')



"""
Copyright 2017-2019 Government of Canada - Public Services and Procurement Canada - buyandsell.gc.ca

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


from urllib.parse import urlparse

from von_anchor.error import BadIdentifier
from von_anchor.util import ok_did


def resource(ref: str, delimiter: str = None) -> str:
    """
    Given a (URI) reference, return up to its delimiter (exclusively), or all of it if there is none.

    :param ref: reference
    :param delimiter: delimiter character (default None maps to '#', or ';' introduces identifiers)
    """

    return ref.split(delimiter if delimiter else '#')[0]


def canon_did(uri: str) -> str:
    """
    Convert a URI into a DID if need be, left-stripping 'did:sov:' if present.
    Return input if already a DID. Raise BadIdentifier for invalid input.

    :param uri: input URI or DID
    :return: corresponding DID
    """

    if ok_did(uri):
        return uri

    if uri.startswith('did:sov:'):
        rv = uri[8:]
        if ok_did(rv):
            return rv
    raise BadIdentifier('Bad specification {} does not correspond to a sovrin DID'.format(uri))


def canon_ref(did: str, ref: str, delimiter: str = None):
    """
    Given a reference in a DID document, return it in its canonical form of a URI.

    :param did: DID acting as the identifier of the DID document
    :param ref: reference to canonicalize, either a DID or a fragment pointing to a location in the DID doc
    :param delimiter: delimiter character marking fragment (default '#') or
        introducing identifier (';') against DID resource
    """

    if not ok_did(did):
        raise BadIdentifier('Bad DID {} cannot act as DID document identifier'.format(did))

    if ok_did(ref):  # e.g., LjgpST2rjsoxYegQDRm7EL
        return 'did:sov:{}'.format(did)

    if ok_did(resource(ref, delimiter)):  # e.g., LjgpST2rjsoxYegQDRm7EL#keys-1
        return 'did:sov:{}'.format(ref)

    if ref.startswith('did:sov:'):  # e.g., did:sov:LjgpST2rjsoxYegQDRm7EL, did:sov:LjgpST2rjsoxYegQDRm7EL#3
        rv = ref[8:]
        if ok_did(resource(rv, delimiter)):
            return ref
        raise BadIdentifier('Bad URI {} does not correspond to a sovrin DID'.format(ref))

    if urlparse(ref).scheme:  # e.g., https://example.com/messages/8377464
        return ref

    return 'did:sov:{}{}{}'.format(did, delimiter if delimiter else '#', ref)  # e.g., 3



"""
Copyright 2017-2019 Government of Canada - Public Services and Procurement Canada - buyandsell.gc.ca

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


import json
import logging
import re

from os import chdir, getcwd, makedirs, readlink, symlink, walk
from os.path import basename, dirname, isfile, islink, join

from indy import blob_storage

from von_anchor.error import AbsentTails, BadIdentifier
from von_anchor.util import (
    B58,
    ok_did,
    ok_cred_def_id,
    ok_rev_reg_id,
    rev_reg_id,
    rev_reg_id2cred_def_id,
    rev_reg_id2tag)


LOGGER = logging.getLogger(__name__)


class Tails:
    """
    Abstraction layer to manage tails files for Issuers and HolderProvers. Uses symbolic link
    to retain association between tails file and corresponding revocation registry identifier.
    """

    MAX_SIZE = 100000

    def __init__(self, base_dir: str, cd_id: str, tag: str = None):
        """
        Initialize programmatic association between revocation registry identifier
        (on credential definition on input identifier plus tag, default most recent),
        and tails file, via symbolic link.

        Raise AbsentTails if (rev reg id) symbolic link or (tails hash) tails file not present.

        :param base_dir: base directory for tails files, thereafter split by cred def id
        :param cd_id: credential definition identifier of interest
        :param tag: revocation registry identifier tag of interest, default to most recent
        """

        LOGGER.debug('Issuer.__init__ >>> base_dir: %s, cd_id: %s, tag: %s', base_dir, cd_id, tag)

        if not ok_cred_def_id(cd_id):
            LOGGER.debug('Tails.__init__ <!< Bad cred def id %s', cd_id)
            raise BadIdentifier('Bad cred def id {}'.format(cd_id))

        if tag is None:
            self._rr_id = Tails.current_rev_reg_id(base_dir, cd_id)
        else:  # including tag == 0
            self._rr_id = rev_reg_id(cd_id, tag)
            if self._rr_id not in [basename(f) for f in Tails.links(base_dir)]:
                LOGGER.debug(
                    'Tails.__init__ <!< No tails file present for cred def id %s on rev reg id tag %s',
                    cd_id,
                    tag)
                raise AbsentTails('No tails file present for cred def id {} on rev reg id tag {}'.format(cd_id, tag))

        path_link = join(Tails.dir(base_dir, self._rr_id), self._rr_id)
        if not islink(path_link):
            raise AbsentTails('No symbolic link present at {} for rev reg id {}'.format(path_link, self._rr_id))

        path_tails = Tails.linked(base_dir, self._rr_id)
        if not isfile(path_tails):
            raise AbsentTails('No tails file present at {} for rev reg id {}'.format(path_tails, self._rr_id))

        self._tails_config_json = json.dumps({
            'base_dir': dirname(path_tails),
            'uri_pattern': '',
            'file': basename(path_tails)
        })

        self._reader_handle = None

        LOGGER.debug('Tails.__init__ <<<')

    async def open(self) -> 'Tails':
        """
        Open reader handle and return current object.

        :return: current object
        """

        LOGGER.debug('Tails.open >>>')

        self._reader_handle = await blob_storage.open_reader('default', self._tails_config_json)

        LOGGER.debug('Tails.open <<<')
        return self

    @staticmethod
    def ok_hash(token: str) -> bool:
        """
        Whether input token looks like a valid tails hash.

        :param token: candidate string
        :return: whether input token looks like a valid tails hash
        """

        LOGGER.debug('Tails.ok_hash >>> token: %s', token)

        rv = re.match('[{}]{{42,44}}$'.format(B58), token) is not None
        LOGGER.debug('Tails.ok_hash <<< %s', rv)
        return rv

    @staticmethod
    def associate(base_dir: str, rr_id: str, tails_hash: str) -> None:
        """
        Create symbolic link to tails file named tails_hash for rev reg id rr_id.

        :param rr_id: rev reg id
        :param tails_hash: hash of tails file, serving as file name
        """

        LOGGER.debug('Tails.associate >>> base_dir: %s, rr_id: %s, tails_hash: %s', base_dir, rr_id, tails_hash)

        if not ok_rev_reg_id(rr_id):
            LOGGER.debug('Tails.associate <!< Bad rev reg id %s', rr_id)
            raise BadIdentifier('Bad rev reg id {}'.format(rr_id))

        if not Tails.ok_hash(tails_hash):
            LOGGER.debug('Tails.associate <!< Bad tails hash %s', tails_hash)
            raise BadIdentifier('Bad tails hash {}'.format(tails_hash))

        cd_id = rev_reg_id2cred_def_id(rr_id)
        directory = join(base_dir, cd_id)
        cwd = getcwd()
        makedirs(directory, exist_ok=True)
        chdir(directory)
        symlink(tails_hash, rr_id)
        chdir(cwd)

        LOGGER.debug('Tails.associate <<<')

    @staticmethod
    def dir(base_dir: str, rr_id: str) -> str:
        """
        Return correct subdirectory of input base dir for artifacts corresponding to input rev reg id.

        :param base_dir: base directory for tails files, thereafter split by cred def id
        :param rr_id: rev reg id
        """

        LOGGER.debug('Tails.dir >>> base_dir: %s, rr_id: %s', base_dir, rr_id)

        if not ok_rev_reg_id(rr_id):
            LOGGER.debug('Tails.dir <!< Bad rev reg id %s', rr_id)
            raise BadIdentifier('Bad rev reg id {}'.format(rr_id))

        rv = join(base_dir, rev_reg_id2cred_def_id(rr_id))
        LOGGER.debug('Tails.dir <<< %s', rv)
        return rv

    @staticmethod
    def linked(base_dir: str, rr_id: str) -> str:
        """
        Get, from the specified directory, the path to the tails file associated with
        the input revocation registry identifier, or None for no such file.

        :param base_dir: base directory for tails files, thereafter split by cred def id
        :param rr_id: rev reg id
        :return: (stringified) path to tails file of interest, or None for no such file.
        """

        LOGGER.debug('Tails.linked >>> base_dir: %s, rr_id: %s', base_dir, rr_id)

        if not ok_rev_reg_id(rr_id):
            LOGGER.debug('Tails.linked <!< Bad rev reg id %s', rr_id)
            raise BadIdentifier('Bad rev reg id {}'.format(rr_id))

        cd_id = rev_reg_id2cred_def_id(rr_id)
        link = join(base_dir, cd_id, rr_id)

        rv = join(base_dir, cd_id, readlink(link)) if islink(link) else None
        LOGGER.debug('Tails.linked <<< %s', rv)
        return rv

    @staticmethod
    def links(base_dir: str, issuer_did: str = None) -> set:
        """
        Return set of all paths to symbolic links (rev reg ids) associating their
        respective tails files, in specified base tails directory recursively
        (omitting the .hopper subdirectory), on input issuer DID if specified.

        :param base_dir: base directory for tails files, thereafter split by cred def id
        :param issuer_did: issuer DID of interest
        :return: set of paths to symbolic links associating tails files
        """

        LOGGER.debug('Tails.links >>> base_dir: %s, issuer_did: %s', base_dir, issuer_did)

        if issuer_did and not ok_did(issuer_did):
            LOGGER.debug('Tails.links <!< Bad DID %s', issuer_did)
            raise BadIdentifier('Bad DID {}'.format(issuer_did))

        rv = set()
        for dir_path, dir_names, file_names in walk(base_dir, topdown=True):
            dir_names[:] = [d for d in dir_names if not d.startswith('.')]
            for file_name in file_names:
                if islink(join(dir_path, file_name)) and (not issuer_did or ok_rev_reg_id(file_name, issuer_did)):
                    rv.add(join(dir_path, file_name))

        LOGGER.debug('Tails.links <<< %s', rv)
        return rv

    @staticmethod
    def unlinked(base_dir: str) -> set:
        """
        Return all paths to tails files, in specified tails base directory recursively
        (omitting the .hopper subdirectory), without symbolic links associating
        revocation registry identifiers.

        At an Issuer, tails files should not persist long without revocation registry identifier
        association via symbolic link. At a HolderProver, a newly downloaded tails file stays
        unlinked until the anchor stores a credential or creates a proof needing it, or else the
        anchor restarts.

        :param base_dir: base directory for tails files, thereafter split by cred def id
        :return: set of paths to tails files with no local symbolic links to them
        """

        LOGGER.debug('Tails.unlinked >>> base_dir: %s', base_dir)

        rv = set()
        for dir_path, dir_names, file_names in walk(base_dir, topdown=True):
            dir_names[:] = [d for d in dir_names if not d.startswith('.')]
            for file_name in file_names:
                if isfile(join(dir_path, file_name)) and Tails.ok_hash(file_name):
                    rv.add(join(dir_path, file_name))
        rv -= {join(dirname(path_link), readlink(path_link)) for path_link in Tails.links(base_dir)}

        LOGGER.debug('Tails.unlinked <<< %s', rv)
        return rv

    @staticmethod
    def next_tag(base_dir: str, cd_id: str) -> (str, int):
        """
        Return the next tag name available for a new rev reg id on input cred def id in base directory,
        and suggested size of associated rev reg.

        :param base_dir: base directory for tails files, thereafter split by cred def id
        :param cd_id: credential definition identifier of interest
        :return: stringified least non-negative integer not yet used in a rev reg id associated with a tails file
            in base directory, and recommendation for next size to use
        """

        LOGGER.debug('Tails.next_tag >>> base_dir: %s, cd_id: %s', base_dir, cd_id)

        if not ok_cred_def_id(cd_id):
            LOGGER.debug('Tails.next_tag <!< Bad cred def id %s', cd_id)
            raise BadIdentifier('Bad cred def id {}'.format(cd_id))

        # append [-1] first to max existing tags: next tag is '0' if no tags so far
        tag = 1 + max([int(rev_reg_id2tag(basename(f))) for f in Tails.links(base_dir) if cd_id in basename(f)] + [-1])
        size = min(2**(tag + 6), Tails.MAX_SIZE)

        rv = (tag, size)
        LOGGER.debug('Tails.next_tag <<< %s', rv)
        return rv

    @staticmethod
    def current_rev_reg_id(base_dir: str, cd_id: str) -> str:
        """
        Return the current revocation registry identifier for
        input credential definition identifier, in input directory.

        Raise AbsentTails if no corresponding tails file, signifying no such revocation registry defined.

        :param base_dir: base directory for tails files, thereafter split by cred def id
        :param cd_id: credential definition identifier of interest
        :return: identifier for current revocation registry on input credential definition identifier
        """

        LOGGER.debug('Tails.current_rev_reg_id >>> base_dir: %s, cd_id: %s', base_dir, cd_id)

        if not ok_cred_def_id(cd_id):
            LOGGER.debug('Tails.current_rev_reg_id <!< Bad cred def id %s', cd_id)
            raise BadIdentifier('Bad cred def id {}'.format(cd_id))

        tags = [int(rev_reg_id2tag(basename(f))) for f in Tails.links(base_dir) if cd_id in basename(f)]
        if not tags:
            raise AbsentTails('No tails files present for cred def id {}'.format(cd_id))

        rv = rev_reg_id(cd_id, str(max(tags)))  # ensure 10 > 9, not '9' > '10'
        LOGGER.debug('Tails.current_rev_reg_id <<< %s', rv)
        return rv

    @property
    def reader_handle(self) -> int:
        """
        Accessor for reader handle on current tails file.
        Note that the current object must be open() for the handle to exist.

        :return: reader handle
        """

        return self._reader_handle

    @property
    def rr_id(self) -> str:
        """
        Accessor for revocation registry identifier.

        :return: revocation registry identifier for current tails file
        """

        return self._rr_id

    @property
    def path(self) -> str:
        """
        Accessor for (stringified) path to current tails file.

        :return: (stringified) path to current tails file.
        """

        config = json.loads(self._tails_config_json)
        return join(config['base_dir'], config['file'])

    def __str__(self) -> str:
        """
        Return string representation.

        :return: string representation
        """

        config = json.loads(self._tails_config_json)
        return 'Tails: {}/{} -> {}'.format(config['base_dir'], self._rr_id, config['file'])



"""
Copyright 2017-2019 Government of Canada - Public Services and Procurement Canada - buyandsell.gc.ca

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


import json
import logging
import re

from asyncio import sleep

from indy import anoncreds, ledger

from von_anchor.anchor.base import BaseAnchor
from von_anchor.cache import SCHEMA_CACHE
from von_anchor.error import AbsentSchema, BadAttribute, BadLedgerTxn
from von_anchor.util import schema_id, schema_key


LOGGER = logging.getLogger(__name__)


class Origin(BaseAnchor):
    """
    Mixin for anchor to send schemata to the distributed ledger.
    """

    async def send_schema(self, schema_data_json: str) -> str:
        """
        Send schema to ledger, then retrieve it as written to the ledger and return it.
        Raise BadLedgerTxn on failure. Raise BadAttribute for attribute name with spaces or
        reserved for indy-sdk.

        If schema already exists on ledger, log error and return schema.

        :param schema_data_json: schema data json with name, version, attribute names; e.g.,

        ::

            {
                'name': 'my-schema',
                'version': '1.234',
                'attr_names': ['favourite_drink', 'height', 'last_visit_date']
            }

        :return: schema json as written to ledger (or existed a priori)
        """

        LOGGER.debug('Origin.send_schema >>> schema_data_json: %s', schema_data_json)

        schema_data = json.loads(schema_data_json)
        for attr in schema_data['attr_names']:
            if not (re.match(r'(?=[^- ])[-_a-zA-Z0-9 ]+(?<=[^- ])$', attr)) or attr.strip().lower() == 'hash':
                LOGGER.debug('Origin.send_schema <!< Bad attribute name [%s]', attr)
                raise BadAttribute('Bad attribute name [{}]'.format(attr))

        s_id = schema_id(self.did, schema_data['name'], schema_data['version'])
        s_key = schema_key(s_id)
        rv_json = None
        with SCHEMA_CACHE.lock:
            try:
                rv_json = await self.get_schema(s_key)
                LOGGER.info(
                    'Schema %s version %s already exists on ledger for origin-did %s: not sending',
                    schema_data['name'],
                    schema_data['version'],
                    self.did)
            except AbsentSchema:  # OK - about to create and send it
                (_, schema_json) = await anoncreds.issuer_create_schema(
                    self.did,
                    schema_data['name'],
                    schema_data['version'],
                    json.dumps(schema_data['attr_names']))
                req_json = await ledger.build_schema_request(self.did, schema_json)
                await self._sign_submit(req_json)

                for _ in range(16):  # reasonable timeout
                    try:
                        rv_json = await self.get_schema(s_key)  # adds to cache
                        break
                    except AbsentSchema:
                        await sleep(1)
                        LOGGER.info('Sent schema %s to ledger, waiting 1s for its appearance', s_id)

                if not rv_json:
                    LOGGER.debug('Origin.send_schema <!< timed out waiting on sent schema %s', s_id)
                    raise BadLedgerTxn('Timed out waiting on sent schema {}'.format(s_id))

        LOGGER.debug('Origin.send_schema <<< %s', rv_json)
        return rv_json



"""
Copyright 2017-2019 Government of Canada - Public Services and Procurement Canada - buyandsell.gc.ca

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


import json
import re

from typing import Any

from von_anchor.error import BadWalletQuery


WQL_1_OPS = frozenset(('$in', '$neq', '$gt', '$gte', '$lt', '$lte', '$like'))


def raw(orig: Any) -> dict:
    """
    Stringify input value, empty string for None.

    :param orig: original attribute value of any stringifiable type
    :return: stringified raw value
    """

    return '' if orig is None else str(orig)


def canon(raw_attr_name: str) -> str:
    """
    Canonicalize input attribute name as it appears in proofs and credential offers: strip out
    white space and convert to lower case.

    :param raw_attr_name: attribute name
    :return: canonicalized attribute name
    """

    if raw_attr_name:  # do not dereference None, and '' is already canonical
        return raw_attr_name.replace(' ', '').lower()
    return raw_attr_name


def canon_cred_wql(query: dict) -> dict:
    """
    Canonicalize WQL attribute marker and value keys for input to indy-sdk wallet credential filtration.
    Canonicalize comparison values to proper indy-sdk raw values as per raw().

    Raise BadWalletQuery for WQL mapping '$or' to non-list.

    :param query: WQL query
    :return: canonicalized WQL query dict
    """

    for k in [qk for qk in query]:  # copy: iteration alters query keys
        attr_match = re.match('attr::([^:]+)::(marker|value)$', k)
        if isinstance(query[k], dict):  # only subqueries are dicts: recurse
            query[k] = canon_cred_wql(query[k])
        if k == '$or':
            if not isinstance(query[k], list):
                raise BadWalletQuery('Bad WQL; $or value must be a list in {}'.format(json.dumps(query)))
            query[k] = [canon_cred_wql(subq) for subq in query[k]]
        if attr_match:
            qkey = 'attr::{}::{}'.format(canon(attr_match.group(1)), canon(attr_match.group(2)))
            query[qkey] = query.pop(k)
            tag_value = query[qkey]
            if isinstance(tag_value, dict) and len(tag_value) == 1:
                if '$in' in tag_value:
                    tag_value['$in'] = [raw(val) for val in tag_value.pop('$in')]
                else:
                    wql_op = set(tag_value.keys()).pop()  # $neq, $gt, $gte, etc.
                    tag_value[wql_op] = raw(tag_value[wql_op])
            else:  # equality
                query[qkey] = raw(query[qkey])

    return query


def canon_non_secret_wql(query: dict = None) -> dict:
    """
    Canonicalize comparison values to strings via raw().

    Raise BadWalletQuery for WQL mapping '$or' to non-list.

    :param query: WQL query
    :return: canonicalized WQL query dict
    """

    if not query:
        return {}

    for k in [qk for qk in query]:  # copy: iteration alters query keys
        if isinstance(query[k], dict):  # only subqueries are dicts: recurse
            query[k] = canon_non_secret_wql(query[k])
        if k == '$or':
            if not isinstance(query[k], list):
                raise BadWalletQuery('Bad WQL; $or value must be a list in {}'.format(json.dumps(query)))
            query[k] = [canon_non_secret_wql(subq) for subq in query[k]]
        elif k == '$not':
            query[k] = canon_non_secret_wql(query.pop(k))
        elif k not in WQL_1_OPS:
            tag_value = query[k]
            if isinstance(tag_value, dict) and len(tag_value) == 1:
                if '$in' in tag_value:
                    tag_value['$in'] = [raw(val) for val in tag_value['$in']]
                else:
                    wql_op = set(tag_value.keys()).pop()  # $neq, $gt, $gt, etc.
                    tag_value[wql_op] = raw(tag_value[wql_op])
            else:
                query[k] = raw(query.pop(k))

    return query


def canon_pairwise_tag(tag: str) -> str:
    """
    Canonicalize pairwise tag to specify unencrypted storage.

    :param tag: input tag
    :return: tag prefixed with '~' if not already
    """

    return '{}{}'.format('' if str(tag).startswith('~') else '~', tag)


def canon_pairwise_wql(query: dict = None) -> dict:
    """
    Canonicalize WQL tags to unencrypted storage specification.
    Canonicalize comparison values to strings via raw().
    Canonicalize empty query to find all pairwise DID records.

    Raise BadWalletQuery for WQL mapping '$or' to non-list.

    :param query: WQL query
    :return: canonicalized WQL query dict
    """

    if not query:
        return {
            '~their_did': {
                '$neq': ''
            }
        }

    for k in [qk for qk in query]:  # copy: iteration alters query keys
        if isinstance(query[k], dict):  # only subqueries are dicts: recurse
            query[k] = canon_pairwise_wql(query[k])
        if k == '$or':
            if not isinstance(query[k], list):
                raise BadWalletQuery('Bad WQL; $or value must be a list in {}'.format(json.dumps(query)))
            query[k] = [canon_pairwise_wql(subq) for subq in query[k]]
        elif k == '$not':
            query[k] = canon_pairwise_wql(query.pop(k))
        elif k not in WQL_1_OPS:
            qkey = canon_pairwise_tag(k)
            query[qkey] = query.pop(k)
            tag_value = query[qkey]
            if isinstance(tag_value, dict) and len(tag_value) == 1:
                if '$in' in tag_value:
                    tag_value['$in'] = [raw(val) for val in tag_value['$in']]
                else:
                    wql_op = set(tag_value.keys()).pop()  # $neq, $gt, $gt, etc.
                    tag_value[wql_op] = raw(tag_value[wql_op])
            else:
                query[qkey] = raw(query.pop(qkey))

    return query



"""
Copyright 2017-2019 Government of Canada - Public Services and Procurement Canada - buyandsell.gc.ca

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


import logging

from os.path import basename

from von_anchor.anchor.base import BaseAnchor
from von_anchor.anchor.holderprover import HolderProver
from von_anchor.anchor.issuer import Issuer
from von_anchor.anchor.origin import Origin
from von_anchor.anchor.smith import AnchorSmith
from von_anchor.anchor.verifier import Verifier
from von_anchor.cache import ArchivableCaches
from von_anchor.error import WalletState
from von_anchor.indytween import Role
from von_anchor.nodepool import NodePool
from von_anchor.tails import Tails
from von_anchor.validcfg import validate_config
from von_anchor.wallet import Wallet


LOGGER = logging.getLogger(__name__)


class NominalAnchor(BaseAnchor):
    """
    NominalAnchor demonstrator class needs access to ledger for authenticated encryption and decryption.
    """

    @staticmethod
    def least_role() -> Role:
        """
        Return the indy-sdk null role for a tails sync anchor, which does not need write access.

        :return: USER role
        """

        LOGGER.debug('NominalAnchor.least_role >>>')

        rv = Role.USER

        LOGGER.debug('NominalAnchor.least_role <<< %s', rv)
        return rv


class TrusteeAnchor(AnchorSmith):
    """
    TrusteeAnchor demonstrator class acts as an anchor smith to forge new anchors.
    """


class RegistrarAnchor(Origin, Issuer):
    """
    RegistrarAnchor demonstrator class acts as an issuer.
    """


class OrgBookAnchor(HolderProver):
    """
    Basic OrgBookAnchor demonstrator class acts as a holder-prover for any of its registrars' credentials.
    """


class OrgHubAnchor(Verifier, Origin, Issuer, OrgBookAnchor):
    """
    OrgHubAnchor demonstrator class acts as an origin, issuer and verifier for its own
    schemata, credentials, and presentations; and as a holder-prover for its own and any
    of its registrars' credentials.
    """

    def __init__(self, wallet: Wallet, pool: NodePool = None, **kwargs) -> None:
        """
        Initializer for org hub anchor. Retain input parameters; do not open wallet.

        :param wallet: wallet for anchor use
        :param pool: pool for anchor use
        :param config: configuration dict for cache archive behaviour qua Issuer and Verifier roles; e.g.,

        ::

            {
                'parse-caches-on-open': True,
                'archive-holder-prover-caches-on-close': True,
                'archive-verifier-caches-on-close': {
                    'schema_id': [
                        'R17v42T4pk...:2:tombstone:1.2',
                        '9cHbp54C8n...:2:business:2.0',
                        'Pcq76cx6jE...:2:birth_cert:1.0',
                        ...
                    ],
                    'cred_def_id': [
                        'R17v42T4pk...:3:CL:19:tag',
                        '9cHbp54C8n...:3:CL:37:tag',
                        'Pcq76cx6jE...:3:CL:51:tag',
                        ...
                    ]
                    'rev_reg_id': [
                        'R17v42T4pk...:4:R17v42T4pk...:3:CL:19:tag:CL_ACCUM:0',
                        'R17v42T4pk...:4:R17v42T4pk...:3:CL:19:tag:CL_ACCUM:1',
                        '9cHbp54C8n...:4:9cHbp54C8n...:3:CL:37:tag:CL_ACCUM:0',
                        '9cHbp54C8n...:4:9cHbp54C8n...:3:CL:37:tag:CL_ACCUM:1',
                        '9cHbp54C8n...:4:9cHbp54C8n...:3:CL:37:tag:CL_ACCUM:2',
                        ...
                    ]
                }
            }

        :param rrbx: whether revocation registry builder is an external process (default False)
        """

        LOGGER.debug('OrgHubAnchor.__init__ >>> wallet: %s, pool: %s, kwargs: %s', wallet, pool, kwargs)

        super().__init__(wallet, pool, **kwargs)

        self._config = kwargs.get('config', {})
        validate_config('org-hub', self._config)

        LOGGER.debug('OrgHubAnchor.__init__ <<<')

    @staticmethod
    def least_role() -> Role:
        """
        Return the indy-sdk TRUST_ANCHOR role for Org Hub anchor.

        :return: TRUST_ANCHOR role
        """

        LOGGER.debug('OrgHubAnchor.least_role >>>')

        rv = Role.TRUST_ANCHOR

        LOGGER.debug('OrgHubAnchor.least_role <<< %s', rv)
        return rv

    async def close(self) -> None:
        """
        Explicit exit. If so configured, populate cache to prove for any creds on schemata,
        cred defs, and rev regs marked of interest in configuration at initialization,
        archive cache, and purge prior cache archives.

        :return: current object
        """

        LOGGER.debug('OrgHubAnchor.close >>>')

        archive_caches = False
        if self.config.get('archive-holder-prover-caches-on-close', False):
            archive_caches = True
            try:
                await self.load_cache_for_proof(False)
            except WalletState:
                LOGGER.warning(
                    'OrgHubAnchor load cache for proof on close required open wallet %s but it was closed',
                    self.name)
        if self.config.get('archive-verifier-caches-on-close', {}):
            archive_caches = True
            await self.load_cache_for_verification(False)
        if archive_caches:
            ArchivableCaches.archive(self.dir_cache)
            ArchivableCaches.purge_archives(self.dir_cache, True)

        # Do not close wallet independently: allow for sharing open wallet over many anchor lifetimes
        # Do not close pool independently: let relying party decide when to go on-line and off-line

        for path_rr_id in Tails.links(self._dir_tails):
            rr_id = basename(path_rr_id)
            await HolderProver._sync_revoc_for_proof(self, rr_id)  # warns for closed pool

        LOGGER.debug('OrgHubAnchor.close <<<')


class ProctorAnchor(Verifier, Origin, Issuer):
    """
    ProctorAnchor demonstrator class acts as an origin and issuer of its own credentials and a verifier
    of a holder-prover's presentations.
    """

    def __init__(self, wallet: Wallet, pool: NodePool = None, **kwargs) -> None:
        """
        Initializer for SRI anchor. Retain input parameters; do not open wallet.

        :param wallet: wallet for anchor use
        :param pool: pool for anchor use
        :param config: configuration dict for cache archive behaviour qua Verifier role
        :param rrbx: whether revocation registry builder is an external process
        """

        LOGGER.debug('ProctorAnchor.__init__ >>> wallet: %s, pool: %s, kwargs: %s', wallet, pool, kwargs)

        super().__init__(wallet, pool, **kwargs)

        LOGGER.debug('ProctorAnchor.close <<<')

    @staticmethod
    def least_role() -> Role:
        """
        Return the indy-sdk TRUST_ANCHOR role for SRI anchor.

        :return: TRUST_ANCHOR role
        """

        LOGGER.debug('ProctorAnchor.least_role >>>')

        rv = Role.TRUST_ANCHOR

        LOGGER.debug('ProctorAnchor.least_role <<< %s', rv)
        return rv



"""
Copyright 2017-2019 Government of Canada - Public Services and Procurement Canada - buyandsell.gc.ca

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


import json
import logging

from os import remove
from os.path import expanduser, expandvars, isfile, realpath
from tempfile import NamedTemporaryFile
from typing import List

from indy import pool
from indy.error import IndyError

from von_anchor.error import AbsentPool, ExtantPool
from von_anchor.nodepool.nodepool import NodePool
from von_anchor.nodepool.protocol import Protocol


LOGGER = logging.getLogger(__name__)


class NodePoolManager:
    """
    Class for managing indy node pools.
    """

    def __init__(self, protocol: Protocol = None):
        """
        Initializer for node pool manager. Set protocol.
        """

        LOGGER.debug('NodePoolManager.__init__ >>> protocol %s', protocol)

        self._protocol = protocol or Protocol.DEFAULT

        LOGGER.debug('NodePoolManager.__init__ <<<')

    @property
    def protocol(self) -> str:
        """
        Accessor for protocol.

        :return: current node manager protocol
        """

        return self._protocol

    @protocol.setter
    def protocol(self, value: Protocol) -> None:
        """
        Set protocol

        :param value: protocol
        """

        self._protocol = value

    async def add_config(self, name: str, genesis: str = None) -> None:
        """
        Given pool name and genesis transaction path or data, add node pool
        configuration to indy home directory.

        Raise ExtantPool if node pool configuration on input name already exists.
        Raise AbsentPool if unable to create the pool ledger configuration.

        :param name: pool name
        :param genesis: genesis transaction path or raw data
        """

        LOGGER.debug('NodePoolManager.add_config >>> name: %s, genesis: %s', name, genesis)

        if name in await self.list():
            LOGGER.debug('NodePoolManager.add_config: <!< Node pool %s configuration already present', name)
            raise ExtantPool('Node pool {} configuration already present'.format(name))

        genesis_tmp = None
        path_gen = realpath(expanduser(expandvars(genesis)))
        try:
            if not isfile(path_gen):
                genesis_tmp = NamedTemporaryFile(mode='w+b', buffering=0, delete=False)
                with genesis_tmp:
                    genesis_tmp.write(genesis.encode())
            await pool.create_pool_ledger_config(
                name,
                json.dumps({
                    'genesis_txn': path_gen if isfile(path_gen) else genesis_tmp.name
                }))
        except IndyError as x_indy:
            LOGGER.debug(
                'NodePoolManager.add_config <!< could not create pool %s ledger configuration: indy error %s',
                name,
                x_indy.error_code)
            raise AbsentPool('Could not create pool {} ledger configuration: indy error {}'.format(
                name,
                x_indy.error_code))

        finally:
            if genesis_tmp:
                remove(genesis_tmp.name)

        LOGGER.debug('NodePoolManager.add_config <<<')

    async def list(self) -> List[str]:
        """
        Return list of pool names configured, empty list for none.

        :return: list of pool names.
        """

        LOGGER.debug('NodePoolManager.list >>>')

        rv = [p['pool'] for p in await pool.list_pools()]

        LOGGER.debug('NodePoolManager.list <<< %s', rv)
        return rv

    def get(self, name: str, config: dict = None) -> NodePool:
        """
        Return node pool in input name and optional configuration.

        :param name: name of configured pool
        :param config: pool configuration with optional 'timeout' int, 'extended_timeout' int,
            'preordered_nodes' array of strings
        :return: node pool
        """

        LOGGER.debug('NodePoolManager.node_pool >>>')

        rv = NodePool(name, self.protocol, config)

        LOGGER.debug('NodePoolManager.node_pool <<< %s', rv)
        return rv

    async def remove(self, name: str) -> None:
        """
        Remove serialized pool info if it exists. Abstain from removing open node pool.
        Raise ExtantPool if deletion fails.
        """

        LOGGER.debug('NodePoolManager.remove >>> name: %s', name)

        try:
            await pool.delete_pool_ledger_config(name)
        except IndyError as x_indy:
            LOGGER.debug(
                'NodePoolManager.remove <!< could not remove %s ledger configuration: indy error %s',
                name,
                x_indy.error_code)
            raise ExtantPool('Could not remove {} ledger configuration: indy error {}'.format(
                name,
                x_indy.error_code))

        LOGGER.debug('NodePool.remove <<<')



"""
Copyright 2017-2019 Government of Canada - Public Services and Procurement Canada - buyandsell.gc.ca

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


import json
import logging

from typing import Sequence
from indy import non_secrets

from von_anchor.canon import canon_non_secret_wql
from von_anchor.error import BadSearch, WalletState
from von_anchor.wallet.wallet import Wallet
from von_anchor.wallet.record import StorageRecord

LOGGER = logging.getLogger(__name__)


class StorageRecordSearch:
    """
    Interactive batch-wise (non-secret) storage record search.
    """

    OPTIONS_JSON = json.dumps({
        'retrieveRecords': True,
        'retrieveTotalCount': True,
        'retrieveType': True,
        'retrieveValue': True,
        'retrieveTags': True
    })

    def __init__(self, wallet: Wallet, typ: str, query: dict):
        """
        Initialize search instance.

        :param wallet: wallet to search
        :param typ: type marking (non-secret) storage records to search
        :param query: WQL query
        """

        LOGGER.debug('StorageRecordSearch.__init__ >>> wallet: %s, typ: %s, query: %s', wallet, typ, query)

        self._wallet = wallet
        self._type = typ
        self._query_json = json.dumps(canon_non_secret_wql(query))

        self._handle = None

        LOGGER.debug('StorageRecordSearch.__init__ <<<')

    @property
    def handle(self) -> int:
        """
        Accessor for search handle.

        :return: search handle
        """

        return self._handle

    @property
    def opened(self) -> bool:
        """
        Accessor for search state.

        :return: whether search is open
        """

        return self._handle is not None

    async def __aenter__(self) -> 'StorageRecordSearch':
        """
        Context manager entry. Open search, for closure on context manager exit.

        :return: current object
        """

        LOGGER.debug('StorageRecordSearch.__aenter__ >>>')

        rv = await self.open()
        LOGGER.debug('StorageRecordSearch.__aenter__ <<<')
        return rv

    async def open(self) -> None:
        """
        Begin the search operation.
        """

        LOGGER.debug('StorageRecordSearch.open >>>')

        if self.opened:
            LOGGER.debug('StorageRecordSearch.open <!< Search is already opened')
            raise BadSearch('Search is already opened')

        if not self._wallet.opened:
            LOGGER.debug('StorageRecordSearch.open <!< Wallet %s is closed', self._wallet.name)
            raise WalletState('Wallet {} is closed'.format(self._wallet.name))

        self._handle = await non_secrets.open_wallet_search(
            self._wallet.handle,
            self._type,
            self._query_json,
            StorageRecordSearch.OPTIONS_JSON)

        LOGGER.debug('StorageRecordSearch.open <<<')

    async def fetch(self, limit: int = None) -> Sequence[StorageRecord]:
        """
        Fetch next batch of search results.

        Raise BadSearch if search is closed, WalletState if wallet is closed.

        :param limit: maximum number of records to return (default value Wallet.DEFAULT_CHUNK)
        :return: next batch of records found
        """

        LOGGER.debug('StorageRecordSearch.fetch >>> limit: %s', limit)

        if not self.opened:
            LOGGER.debug('StorageRecordSearch.fetch <!< Storage record search is closed')
            raise BadSearch('Storage record search is closed')

        if not self._wallet.opened:
            LOGGER.debug('StorageRecordSearch.fetch <!< Wallet %s is closed', self._wallet.name)
            raise WalletState('Wallet {} is closed'.format(self._wallet.name))

        records = json.loads(await non_secrets.fetch_wallet_search_next_records(
            self._wallet.handle,
            self.handle,
            limit or Wallet.DEFAULT_CHUNK))['records'] or []  # at exhaustion results['records'] = None

        rv = [StorageRecord(typ=rec['type'], value=rec['value'], tags=rec['tags'], ident=rec['id']) for rec in records]
        LOGGER.debug('StorageRecordSearch.fetch <<< %s', rv)
        return rv

    async def __aexit__(self, exc_type, exc, traceback) -> None:
        """
        Context manager exit. Close search.

        :param exc_type:
        :param exc:
        :param traceback:
        """

        LOGGER.debug('StorageRecordSearch.__aexit__ >>>')

        await self.close()

        LOGGER.debug('StorageRecordSearch.__aexit__ <<<')

    async def close(self) -> None:
        """
        Close search.
        """

        LOGGER.debug('StorageRecordSearch.close >>>')

        if self._handle:
            await non_secrets.close_wallet_search(self.handle)
            self._handle = None

        LOGGER.debug('StorageRecordSearch.close <<<')



"""
Copyright 2017-2019 Government of Canada - Public Services and Procurement Canada - buyandsell.gc.ca

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


import logging

from uuid import uuid4

from von_anchor.error import BadRecord


TYPE_LINK_SECRET_LABEL = 'link_secret_label'
TYPE_PAIRWISE = 'pairwise'

LOGGER = logging.getLogger(__name__)


class StorageRecord:
    """
    Non-secret wallet record.
    """

    def __init__(self, typ: str, value: str, tags: dict = None, ident: str = None) -> None:
        """
        Initialize non-secret record. Raise BadRecord if input tags are not legitimate indy non-secret record tags.

        :param typ: record type - (typ, ident) identifies a non-secret record in the wallet
        :param value: record value
        :param tags: record tags (metadata) dict
        :param ident: record identifier - (typ, ident) identifies a non-secret record in the wallet
        """

        self._type = typ
        self._id = ident or uuid4().hex
        self._value = value

        if not StorageRecord.ok_tags(tags):
            LOGGER.debug('StorageRecord.__init__ <!< Tags %s must map strings to strings', tags)
            raise BadRecord('Tags {} must map strings to strings'.format(tags))

        self._tags = tags or {}  # store trivial tags as empty (for iteration), return as None

    @staticmethod
    def ok_tags(tags: dict) -> bool:
        """
        Whether input tags dict is OK as an indy-sdk tags structure (depth=1, string values).
        """

        if not tags:
            return True
        depth = 0
        queue = [(i, depth+1) for i in tags.values() if isinstance(i, dict)]
        max_depth = 0
        while queue and max_depth < 2:
            sub, depth = queue.pop()
            max_depth = max(max_depth, depth)
            queue = queue + [(i, depth+1) for i in sub.values() if isinstance(i, dict)]

        return max_depth < 2 and all(isinstance(k, str) and isinstance(tags[k], str) for k in tags)

    @property
    def type(self) -> str:
        """
        Accessor for record type.

        :return: type
        """

        return self._type

    @property
    def id(self) -> str:
        """
        Accessor for record identifier.

        :return: record identifier
        """

        return self._id

    @id.setter
    def id(self, val: str) -> None:
        """
        Accessor for record identifier.

        :param val: identifier value
        """

        self._id = val

    @property
    def value(self) -> str:
        """
        Accessor for record value.

        :return: record value
        """

        return self._value

    @value.setter
    def value(self, val: str) -> None:
        """
        Accessor for record value.

        :param val: record value
        """

        self._value = val

    @property
    def tags(self) -> dict:
        """
        Accessor for record tags (metadata).

        :return: record tags
        """

        return self._tags or None  # store trivial tags as empty (for iteration), return as None

    @tags.setter
    def tags(self, val: str) -> None:
        """
        Accessor for record tags (metadata).

        :param val: record tags
        """

        if not StorageRecord.ok_tags(val):
            LOGGER.debug('StorageRecord.__init__ <!< Tags %s must map strings to strings', val)
            raise BadRecord('Tags {} must map strings to strings'.format(val))

        self._tags = val or {}

    @property
    def clear_tags(self) -> dict:
        """
        Accessor for record tags (metadata) stored in the clear.

        :return: record tags stored in the clear
        """

        return {t: self.tags[t] for t in (self.tags or {}) if t.startswith('~')} or None

    @property
    def encr_tags(self) -> dict:
        """
        Accessor for record tags (metadata) stored encrypted.

        :return: record tags stored encrypted
        """

        return {t: self._tags[t] for t in self.tags or {} if not t.startswith('~')} or None

    def __eq__(self, other: 'DIDInfo') -> bool:
        """
        Equivalence operator. Two instances are equivalent when their attributes are.

        :param other: instance to test for equivalence
        :return: whether instances are equivalent
        """

        return self.type == other.type and self.id == other.id and self.value == other.value and self.tags == other.tags

    def __repr__(self) -> str:
        """
        Return representation.

        :return: string representation evaluating to construction call
        """

        return 'StorageRecord({}, {}, {}, {})'.format(self.type, self.id, self.value, self.tags)



"""
Copyright 2017-2019 Government of Canada - Public Services and Procurement Canada - buyandsell.gc.ca

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


import json
import logging

from os import listdir, makedirs
from os.path import basename, expanduser, isdir, isfile, join
from time import time
from typing import Sequence, Union

from indy import anoncreds, ledger
from indy.error import IndyError, ErrorCode

from von_anchor.anchor.base import BaseAnchor
from von_anchor.cache import ArchivableCaches, RevoCacheEntry, CRED_DEF_CACHE, REVO_CACHE, SCHEMA_CACHE
from von_anchor.canon import canon_cred_wql
from von_anchor.error import (
    AbsentCred,
    AbsentCredDef,
    AbsentInterval,
    AbsentLinkSecret,
    AbsentPool,
    AbsentRevReg,
    AbsentSchema,
    AbsentTails,
    BadIdentifier,
    BadRevStateTime,
    CacheIndex,
    ClosedPool,
    CredentialFocus,
    WalletState)
from von_anchor.indytween import Predicate
from von_anchor.nodepool import NodePool
from von_anchor.tails import Tails
from von_anchor.util import (
    cred_def_id2seq_no,
    iter_briefs,
    ok_cred_def_id,
    ok_rev_reg_id,
    ok_schema_id,
    proof_req_pred_referents,
    rev_reg_id2cred_def_id_tag)
from von_anchor.validcfg import validate_config
from von_anchor.wallet import Wallet, WalletManager


LOGGER = logging.getLogger(__name__)


class HolderProver(BaseAnchor):
    """
    Mixin for anchor acting in the capacity of w3c Holder and indy-sdk Prover.  A Holder holds
    credentials; a Prover produces proof of credentials. Revocation support requires
    the holder-prover anchor to manage tails files.
    """

    def __init__(self, wallet: Wallet, pool: NodePool = None, **kwargs) -> None:
        """
        Initializer for HolderProver anchor. Retain input parameters; do not open wallet nor tails writer.

        :param wallet: wallet for anchor use
        :param pool: pool for anchor use
        :param config: configuration dict for cache archive behaviour; e.g.,

        ::

            {
                'parse-caches-on-open': True,
                'archive-holder-prover-caches-on-close': True
            }

        """

        LOGGER.debug('HolderProver.__init__ >>> wallet: %s, pool: %s, kwargs: %s', wallet, pool, kwargs)

        super().__init__(wallet, pool, **kwargs)

        self._dir_tails = join(expanduser('~'), '.indy_client', 'tails')
        makedirs(self._dir_tails, exist_ok=True)

        self._config = kwargs.get('config', {})
        validate_config('holder-prover', self._config)

        self._dir_cache = join(expanduser('~'), '.indy_client', 'cache', self.name)
        makedirs(self._dir_cache, exist_ok=True)

        LOGGER.debug('HolderProver.__init__ <<<')

    async def _assert_link_secret(self, action: str) -> str:
        """
        Return current wallet link secret label. Raise AbsentLinkSecret if link secret is not set.

        :param action: action requiring link secret
        """

        rv = await self.wallet.get_link_secret_label()
        if rv is None:
            LOGGER.debug('HolderProver._assert_link_secret: action %s requires link secret but it is not set', action)
            raise AbsentLinkSecret('Action {} requires link secret but it is not set'.format(action))

        return rv

    @property
    def config(self) -> dict:
        """
        Accessor for configuration dict

        :return: holder-prover config dict
        """

        return self._config

    @config.setter
    def config(self, value: dict) -> None:
        """
        Set configuration dict

        :param value: configuration dict
        """

        validate_config('holder-prover', value or {})
        self._config = value or {}

    @property
    def dir_cache(self) -> str:
        """
        Accessor for cache archive directory

        :return: holder-prover cache archive directory
        """

        return self._dir_cache

    async def _sync_revoc_for_proof(self, rr_id: str) -> None:
        """
        Pick up tails file reader handle for input revocation registry identifier.  If no symbolic
        link is present, get the revocation registry definition to retrieve its tails file hash,
        then find the tails file and link it.

        Raise AbsentTails for missing corresponding tails file.

        :param rr_id: revocation registry identifier
        """

        LOGGER.debug('HolderProver._sync_revoc_for_proof >>> rr_id: %s', rr_id)

        if not ok_rev_reg_id(rr_id):
            LOGGER.debug('HolderProver._sync_revoc_for_proof <!< Bad rev reg id %s', rr_id)
            raise BadIdentifier('Bad rev reg id {}'.format(rr_id))

        (cd_id, tag) = rev_reg_id2cred_def_id_tag(rr_id)

        try:
            json.loads(await self.get_cred_def(cd_id))
        except AbsentCredDef:
            LOGGER.debug(
                'HolderProver._sync_revoc_for_proof <!< corrupt tails tree %s may be for another ledger',
                self._dir_tails)
            raise AbsentCredDef('Corrupt tails tree {} may be for another ledger'.format(self._dir_tails))
        except (AbsentPool, ClosedPool):  # carry on, may be OK from cache only
            LOGGER.warning(
                'HolderProver %s could not sync cred def id %s for proof: pool %s',
                self.name,
                cd_id,
                '{} closed'.format(self.pool.name) if self.pool else 'absent')

        with REVO_CACHE.lock:
            revo_cache_entry = REVO_CACHE.get(rr_id, None)
            tails = revo_cache_entry.tails if revo_cache_entry else None
            if tails is None:  #  it's not yet set in cache
                try:
                    tails = await Tails(self._dir_tails, cd_id, tag).open()
                except AbsentTails:  # get hash from ledger and check for tails file
                    try:
                        rr_def = json.loads(await self.get_rev_reg_def(rr_id))
                        tails_hash = rr_def['value']['tailsHash']
                        path_tails = join(Tails.dir(self._dir_tails, rr_id), tails_hash)
                        if not isfile(path_tails):
                            LOGGER.debug(
                                'HolderProver._sync_revoc_for_proof <!< No tails file present at %s',
                                path_tails)
                            raise AbsentTails('No tails file present at {}'.format(path_tails))
                        Tails.associate(self._dir_tails, rr_id, tails_hash)
                        tails = await Tails(self._dir_tails, cd_id, tag).open()  # OK now since tails file present
                    except (AbsentPool, ClosedPool):  # carry on, may be OK from cache only
                        LOGGER.warning(
                            'HolderProver %s could not sync rev reg id %s for proof: pool %s',
                            self.name,
                            rr_id,
                            '{} closed'.format(self.pool.name) if self.pool else 'absent')

                if tails:
                    if revo_cache_entry is None:
                        REVO_CACHE[rr_id] = RevoCacheEntry(None, tails)
                    else:
                        REVO_CACHE[rr_id].tails = tails

        LOGGER.debug('HolderProver._sync_revoc_for_proof <<<')

    async def _build_rr_delta_json(self, rr_id: str, to: int, fro: int = None, fro_delta: dict = None) -> (str, int):
        """
        Build rev reg delta json, potentially starting from existing (earlier) delta.

        Return delta json and its timestamp on the distributed ledger.

        Raise AbsentRevReg for no such revocation registry, or BadRevStateTime for a requested delta to
        a time preceding revocation registry creation.

        :param rr_id: rev reg id
        :param to: time (epoch seconds) of interest; upper-bounds returned timestamp
        :param fro: optional prior time of known delta json
        :param fro_delta: optional known delta as of time fro
        :return: rev reg delta json and ledger timestamp (epoch seconds)
        """

        LOGGER.debug(
            '_HolderProver._build_rr_delta_json >>> rr_id: %s, to: %s, fro: %s, fro_delta: %s',
            rr_id,
            to,
            fro,
            fro_delta)

        if not ok_rev_reg_id(rr_id):
            LOGGER.debug('HolderProver._build_rr_delta_json <!< Bad rev reg id %s', rr_id)
            raise BadIdentifier('Bad rev reg id {}'.format(rr_id))

        rr_delta_json = None
        ledger_timestamp = None

        get_rr_delta_req_json = await ledger.build_get_revoc_reg_delta_request(self.did, rr_id, fro, to)
        resp_json = await self._submit(get_rr_delta_req_json)
        resp = json.loads(resp_json)
        if resp.get('result', {}).get('data', None) and resp['result']['data'].get('value', None):
            # delta is to some time at or beyond rev reg creation, carry on
            try:
                (_, rr_delta_json, ledger_timestamp) = await ledger.parse_get_revoc_reg_delta_response(resp_json)
            except IndyError:  # ledger replied, but there is no such rev reg
                LOGGER.debug('_HolderProver._build_rr_delta_json <!< no rev reg exists on %s', rr_id)
                raise AbsentRevReg('No rev reg exists on {}'.format(rr_id))
        else:
            LOGGER.debug(
                '_HolderProver._build_rr_delta_json <!< Rev reg %s created after asked-for time %s',
                rr_id,
                to)
            raise BadRevStateTime('Rev reg {} created after asked-for time {}'.format(rr_id, to))

        if fro and fro_delta:
            rr_delta_json = await anoncreds.issuer_merge_revocation_registry_deltas(
                json.dumps(fro_delta),
                rr_delta_json)

        rv = (rr_delta_json, ledger_timestamp)
        LOGGER.debug('_HolderProver._build_rr_delta_json <<< %s', rv)
        return rv

    def dir_tails(self, rr_id: str) -> str:
        """
        Return path to the correct directory for the tails file on input revocation registry identifier.

        :param rr_id: revocation registry identifier of interest
        :return: path to tails dir for input revocation registry identifier
        """

        LOGGER.debug('HolderProver.dir_tails >>>')

        if not ok_rev_reg_id(rr_id):
            LOGGER.debug('HolderProver.dir_tails <!< Bad rev reg id %s', rr_id)
            raise BadIdentifier('Bad rev reg id {}'.format(rr_id))

        rv = Tails.dir(self._dir_tails, rr_id)
        LOGGER.debug('HolderProver.dir_tails <<< %s', rv)
        return rv

    async def open(self) -> 'HolderProver':
        """
        Explicit entry. Perform ancestor opening operations,
        then parse cache from archive if so configured, and
        synchronize revocation registry to tails tree content.

        :return: current object
        """

        LOGGER.debug('HolderProver.open >>>')

        await super().open()
        if self.config.get('parse-caches-on-open', False):
            ArchivableCaches.parse(self.dir_cache)

        for path_rr_id in Tails.links(self._dir_tails):
            await self._sync_revoc_for_proof(basename(path_rr_id))

        LOGGER.debug('HolderProver.open <<<')
        return self

    async def close(self) -> None:
        """
        Explicit exit. If so configured, populate cache to prove all creds in
        wallet offline if need be, archive cache, and purge prior cache archives.

        :return: current object
        """

        LOGGER.debug('HolderProver.close >>>')

        if self.config.get('archive-holder-prover-caches-on-close', False):
            await self.load_cache_for_proof(True)
            ArchivableCaches.purge_archives(self.dir_cache, True)

        await BaseAnchor.close(self)
        for path_rr_id in Tails.links(self._dir_tails):
            rr_id = basename(path_rr_id)
            try:
                await self._sync_revoc_for_proof(rr_id)
            except ClosedPool:
                LOGGER.warning('HolderProver sync-revoc on close required ledger for %s but pool was closed', rr_id)

        LOGGER.debug('HolderProver.close <<<')

    async def rev_regs(self) -> list:
        """
        Return list of revocation registry identifiers for which HolderProver has associated tails files.
        The operation creates associations for any (newly copied, via service wrapper API) tails files without.

        :return: list of revocation registry identifiers for which HolderProver has associated tails files
        """

        LOGGER.debug('HolderProver.rev_regs >>>')

        for path_rr_id in Tails.links(self._dir_tails):
            await self._sync_revoc_for_proof(basename(path_rr_id))

        rv = [basename(f) for f in Tails.links(self._dir_tails)]
        LOGGER.debug('HolderProver.rev_regs <<< %s', rv)
        return rv

    async def offline_intervals(self, cd_ids: Union[str, Sequence[str]]) -> dict:
        """
        Return default non-revocation intervals for input cred def ids, based on content of revocation cache,
        for augmentation into specification for Verifier.build_proof_req_json. Note that the close() call
        to set the anchor off-line extends all revocation cache registry delta entries to its time of execution:
        in this case, the intervals will all be single timestamps rather than (to, fro) pairs.

        Raise CacheIndex if proof request cites credential definition without corresponding
        content in cred def cache or revocation cache.

        :param cd_ids: credential definition identifier or sequence thereof
        :return: dict mapping revocable cred def ids to interval specifications to augment into cd_id2spec
            parameter for Verifier.build_proof_req_json(), and non-revocable cred def ids to empty dict; e.g.,

        ::

            {
                'Vx4E82R17q...:3:CL:16:tag': {
                    'interval': (1528111730, 1528115832)
                },
                'R17v42T4pk...:3:CL:19:tag': {},
                'Z9ccax812j...:3:CL:27:tag': {
                    'interval': (1528112408, 1528116008)
                },
                '9cHbp54C8n...:3:CL:37:tag': {
                    'interval': 1528116426
                },
                '6caBcmLi33...:tag:CL:41:tag': {},
                ...
            }
        """

        LOGGER.debug('HolderProver.offline_intervals >>> cd_ids: %s', cd_ids)

        rv = {}
        for cd_id in [cd_ids] if isinstance(cd_ids, str) else cd_ids:
            if not ok_cred_def_id(cd_id):
                LOGGER.debug('HolderProver.offline_intervals <!< Bad cred def id %s', cd_id)
                raise BadIdentifier('Bad cred def id {}'.format(cd_id))

            try:
                cred_def = json.loads(await self.get_cred_def(cd_id))
            except ClosedPool:
                LOGGER.debug('HolderProver.offline_intervals <!< no such cred def %s in cred def cache', cd_id)
                raise CacheIndex('No such cred def {} in cred def cache'.format(cd_id))

            rv[cd_id] = {}
            if 'revocation' in cred_def['value']:
                with REVO_CACHE.lock:
                    (fro, to) = REVO_CACHE.dflt_interval(cd_id)
                    if not (fro and to):
                        LOGGER.debug(
                            'HolderProver.offline_intervals <!< no cached delta for non-revoc interval on %s',
                            cd_id)
                        raise CacheIndex('No cached delta for non-revoc interval on {}'.format(cd_id))

                    rv[cd_id]['interval'] = to if fro == to else (fro, to)

        LOGGER.debug('HolderProver.offline_intervals <<< %s', rv)
        return rv

    async def create_link_secret(self, label: str) -> None:
        """
        Create link secret (a.k.a. master secret) used in proofs by HolderProver, if the
        current link secret does not already correspond to the input link secret label.

        Raise WalletState if wallet is closed, or any other IndyError causing failure
        to set link secret in wallet.

        :param label: label for link secret; indy-sdk uses label to generate link secret
        """

        LOGGER.debug('HolderProver.create_link_secret >>> label: %s', label)

        await self.wallet.create_link_secret(label)

        LOGGER.debug('HolderProver.create_link_secret <<<')

    async def create_cred_req(self, cred_offer_json: str, cd_id: str, my_did: str = None) -> (str, str):
        """
        Create credential request as HolderProver and store in wallet; return credential json and metadata json.

        Raise WalletState if wallet is closed; raise AbsentLinkSecret if link secret not set.

        :param cred_offer_json: credential offer json
        :param cd_id: credential definition identifier
        :param my_did: HolderProver DID to use in offer creation (default to anchor DID,
            or specify local DID in pairwise relation)
        :return: cred request json and corresponding metadata json as created and stored in wallet
        """

        LOGGER.debug(
            'HolderProver.create_cred_req >>> cred_offer_json: %s, cd_id: %s, my_did %s',
            cred_offer_json,
            cd_id,
            my_did)

        if not ok_cred_def_id(cd_id):
            LOGGER.debug('HolderProver.create_cred_req <!< Bad cred def id %s', cd_id)
            raise BadIdentifier('Bad cred def id {}'.format(cd_id))

        if not self.wallet.handle:
            LOGGER.debug('HolderProver.create_cred_req <!< Wallet %s is closed', self.name)
            raise WalletState('Wallet {} is closed'.format(self.name))

        label = await self._assert_link_secret('create_cred_req')

        # Check that ledger has schema on ledger where cred def expects - in case of pool reset with extant wallet
        cred_def_json = await self.get_cred_def(cd_id)
        schema_seq_no = int(json.loads(cred_def_json)['schemaId'])
        schema_json = await self.get_schema(schema_seq_no)
        schema = json.loads(schema_json)
        if not schema:
            LOGGER.debug(
                'HolderProver.create_cred_req <!< absent schema@#%s, cred req may be for another ledger',
                schema_seq_no)
            raise AbsentSchema('Absent schema@#{}, cred req may be for another ledger'.format(schema_seq_no))
        (cred_req_json, cred_req_metadata_json) = await anoncreds.prover_create_credential_req(
            self.wallet.handle,
            my_did or self.did,
            cred_offer_json,
            cred_def_json,
            label)
        rv = (cred_req_json, cred_req_metadata_json)

        LOGGER.debug('HolderProver.create_cred_req <<< %s', rv)
        return rv

    async def set_cred_attr_tag_policy(
            self,
            cred_def_id: str,
            taggables: Union[str, Sequence[str]],
            retroactive: bool = False) -> None:
        """
        Set credential attribute tagging policy (in the wallet) by credential definition identifier.

        If taggables is None, clear the policy (restore to default, tagging all attributes).
        If taggables is an empty sequence, set the policy to tag no attributes.
        If retroactive, visit all existing credentials and adjust their tags to comply with incoming policy.

        Raise WalletState if wallet is closed.

        :param cred_def_id: credential definition identifier
        :param taggables: attributes to mark taggable on credential storage: empty sequence for none, None for all
        :param retroactive: whether to reset tags for all existing credentials on input cred def id to specified policy.
        """

        LOGGER.debug(
            'HolderProver.set_cred_attr_tag_policy >>> cred_def_id: %s, taggables: %s, retroactive: %s',
            cred_def_id,
            taggables,
            retroactive)

        if not self.wallet.handle:
            LOGGER.debug('HolderProver.set_cred_attr_tag_policy <!< Wallet %s is closed', self.name)
            raise WalletState('Wallet {} is closed'.format(self.name))

        taggables_json = None
        if isinstance(taggables, str):
            taggables_json = json.dumps([taggables])  # singular
        elif isinstance(taggables, list):
            taggables_json = json.dumps(taggables)
        elif taggables is not None:
            taggables_json = json.dumps([t for t in taggables])  # coerce to list for JSON

        await anoncreds.prover_set_credential_attr_tag_policy(
            self.wallet.handle,
            cred_def_id,
            taggables_json,
            retroactive)

        LOGGER.debug('HolderProver.set_cred_attr_tag_policy <<<')

    async def get_cred_attr_tag_policy(self, cred_def_id: str) -> str:
        """
        Get credential attribute tagging policy (in the wallet) as JSON list (JSON null for default)
        by credential definition identifier.

        Raise WalletState if wallet is closed.

        :param cred_def_id: credential definition identifier
        :return: JSON list of attributes that policy marks taggable, JSON null for no policy (all attributes taggable)
        """

        LOGGER.debug('HolderProver.get_cred_attr_tag_policy >>> cred_def_id: %s', cred_def_id)

        if not self.wallet.handle:
            LOGGER.debug('HolderProver.get_cred_attr_tag_policy <!< Wallet %s is closed', self.name)
            raise WalletState('Wallet {} is closed'.format(self.name))

        rv = await anoncreds.prover_get_credential_attr_tag_policy(self.wallet.handle, cred_def_id)

        LOGGER.debug('HolderProver.get_cred_attr_tag_policy <<< %s', rv)
        return rv

    async def store_cred(self, cred_json: str, cred_req_metadata_json: str) -> str:
        """
        Store cred in wallet as HolderProver, return its credential identifier as created in wallet.

        Raise AbsentTails if tails file not available for revocation registry for input credential.
        Raise WalletState if wallet is closed.

        :param cred_json: credential json as HolderProver created
        :param cred_req_metadata_json: credential request metadata json as HolderProver created via create_cred_req()
        :return: credential identifier within wallet
        """

        LOGGER.debug(
            'HolderProver.store_cred >>> cred_json: %s, cred_req_metadata_json: %s',
            cred_json,
            cred_req_metadata_json)

        if not self.wallet.handle:
            LOGGER.debug('HolderProver.store_cred <!< Wallet %s is closed', self.name)
            raise WalletState('Wallet {} is closed'.format(self.name))

        cred = json.loads(cred_json)
        cred_def_json = await self.get_cred_def(cred['cred_def_id'])
        rr_id = cred['rev_reg_id']
        rr_def_json = None
        if rr_id:
            await self._sync_revoc_for_proof(rr_id)
            rr_def_json = await self.get_rev_reg_def(rr_id)

        rv = await anoncreds.prover_store_credential(
            self.wallet.handle,
            None,  # cred_id, let indy-sdk generate random uuid
            cred_req_metadata_json,
            cred_json,
            cred_def_json,
            rr_def_json)

        LOGGER.debug('HolderProver.store_cred <<< %s', rv)
        return rv

    async def load_cache_for_proof(self, archive: bool = False) -> int:
        """
        Load schema, cred def, revocation caches; optionally archive enough to go
        offline and be able to generate proof on all credentials in wallet.

        Return timestamp (epoch seconds) of cache load event, also used as subdirectory
        for cache archives.

        :param archive: True to archive now or False to demur (subclasses may still
            need to augment archivable caches further)
        :return: cache load event timestamp (epoch seconds)
        """

        LOGGER.debug('HolderProver.load_cache_for_proof >>> archive: %s', archive)

        rv = int(time())
        box_ids = json.loads(await self.get_box_ids_held())
        for s_id in box_ids['schema_id']:
            with SCHEMA_CACHE.lock:
                await self.get_schema(s_id)
        for cd_id in box_ids['cred_def_id']:
            with CRED_DEF_CACHE.lock:
                await self.get_cred_def(cd_id)
        for rr_id in box_ids['rev_reg_id']:
            await self.get_rev_reg_def(rr_id)
            with REVO_CACHE.lock:
                revo_cache_entry = REVO_CACHE.get(rr_id, None)
                if revo_cache_entry:
                    try:
                        await revo_cache_entry.get_delta_json(self._build_rr_delta_json, rv, rv)
                    except (AbsentPool, ClosedPool):
                        LOGGER.warning(
                            'HolderProver %s pool %s, cannot update revo cache reg delta for %s to %s',
                            self.name,
                            '{} closed'.format(self.pool.name) if self.pool else 'absent',
                            rr_id,
                            rv)

        if archive:
            ArchivableCaches.archive(self.dir_cache)
        LOGGER.debug('HolderProver.load_cache_for_proof <<< %s', rv)
        return rv

    async def get_box_ids_held(self) -> str:
        """
        Return json object on lists of all unique box identifiers for credentials in wallet, as
        evidenced by tails directory content:

        * schema identifiers
        * credential definition identifiers
        * revocation registry identifiers.

        E.g.,

        ::

            {
                "schema_id": [
                    "R17v42T4pk...:2:tombstone:1.2",
                    "9cHbp54C8n...:2:business:2.0",
                    ...
                ],
                "cred_def_id": [
                    "R17v42T4pk...:3:CL:19:tag",
                    "9cHbp54C8n...:3:CL:37:tag",
                    ...
                ]
                "rev_reg_id": [
                    "R17v42T4pk...:4:R17v42T4pk...:3:CL:19:tag:CL_ACCUM:0",
                    "R17v42T4pk...:4:R17v42T4pk...:3:CL:19:tag:CL_ACCUM:1",
                    "9cHbp54C8n...:4:9cHbp54C8n...:3:CL:37:tag:CL_ACCUM:0",
                    "9cHbp54C8n...:4:9cHbp54C8n...:3:CL:37:tag:CL_ACCUM:1",
                    "9cHbp54C8n...:4:9cHbp54C8n...:3:CL:37:tag:CL_ACCUM:2",
                    ...
                ]
            }

        Raise WalletState if wallet is closed.

        :return: tuple of sets for schema ids, cred def ids, rev reg ids
        """

        LOGGER.debug('HolderProver.get_box_ids_held >>>')

        if not self.wallet.handle:
            LOGGER.debug('HolderProver.get_box_ids_held <!< Wallet %s is closed', self.name)
            raise WalletState('Wallet {} is closed'.format(self.name))

        rr_ids = {basename(link) for link in Tails.links(self._dir_tails)}

        un_rr_ids = set()
        for rr_id in rr_ids:
            if not json.loads(await self.get_cred_infos_by_q(json.dumps({'rev_reg_id': rr_id}), 1)):
                un_rr_ids.add(rr_id)
        rr_ids -= un_rr_ids

        cd_ids = {
            cd_id for cd_id in listdir(self._dir_tails)
            if isdir(join(self._dir_tails, cd_id)) and ok_cred_def_id(cd_id)
        }
        s_ids = set()
        for cd_id in cd_ids:
            s_ids.add(json.loads(await self.get_schema(cred_def_id2seq_no(cd_id)))['id'])

        un_cd_ids = set()
        for cd_id in cd_ids:
            if not json.loads(await self.get_cred_infos_by_q(json.dumps({'cred_def_id': cd_id}), 1)):
                un_cd_ids.add(cd_id)
        cd_ids -= un_cd_ids

        un_s_ids = set()
        for s_id in s_ids:
            if not json.loads(await self.get_cred_infos_by_q(json.dumps({'schema_id': s_id}), 1)):
                un_s_ids.add(s_id)
        s_ids -= un_s_ids

        rv = json.dumps({
            'schema_id': list(s_ids),
            'cred_def_id': list(cd_ids),
            'rev_reg_id': list(rr_ids)
        })
        LOGGER.debug('HolderProver.get_box_ids_held <<< %s', rv)
        return rv

    async def get_cred_infos_by_q(self, query_json: str, limit: int = None) -> str:
        """
        A cred-info aggregates:
        * its wallet cred-id (aka wallet referent)
        * its attribute names and values
        * its schema identifier
        * its credential definition identifier
        * its revocation registry identifier
        * its credential revocation identifier.

        Return list of cred-infos from wallet by input WQL query;
        return cred-infos for all credentials in wallet for no query.

        Raise WalletState if the wallet is closed.

        The operation supports a subset of WQL; i.e.,

        ::

            query = {subquery}
            subquery = {subquery, ..., subquery} - WHERE subquery AND ... AND subquery
            subquery = $or: [{subquery},..., {subquery}] - WHERE subquery OR ... OR subquery
            subquery = $not: {subquery} - Where NOT (subquery)
            subquery = "tagName": tagValue - WHERE tagName == tagValue
            subquery = "tagName": {$in: [tagValue, ..., tagValue]} - WHERE tagName IN (tagValue, ..., tagValue)
            subquery = "tagName": {$neq: tagValue} - WHERE tagName != tagValue

        but not

        ::

            subquery = "tagName": {$gt: tagValue} - WHERE tagName > tagValue
            subquery = "tagName": {$gte: tagValue} - WHERE tagName >= tagValue
            subquery = "tagName": {$lt: tagValue} - WHERE tagName < tagValue
            subquery = "tagName": {$lte: tagValue} - WHERE tagName <= tagValue
            subquery = "tagName": {$like: tagValue} - WHERE tagName LIKE tagValue

        :param query_json: WQL query json
        :param limit: maximum number of results to return

        :return: cred-infos as json list; i.e.,

        ::

            [
                {
                    "referent": string,  # credential identifier in the wallet
                    "attrs": {
                        "attr1" : {"raw": "value1", "encoded": "value1_as_int" },
                        "attr2" : {"raw": "value2", "encoded": "value2_as_int" },
                        ...
                    }
                    "schema_id": string,
                    "cred_def_id": string,
                    "rev_reg_id": Optional<string>,
                    "cred_rev_id": Optional<string>
                },
                ...
            ]

        """

        LOGGER.debug('HolderProver.get_cred_infos_by_q >>> query_json: %s, limit: %s', query_json, limit)

        if not self.wallet.handle:
            LOGGER.debug('HolderProver.get_cred_infos_by_q <!< Wallet %s is closed', self.name)
            raise WalletState('Wallet {} is closed'.format(self.name))

        infos = []
        if limit and limit < 0:
            limit = None

        (handle, cardinality) = await anoncreds.prover_search_credentials(
            self.wallet.handle,
            json.dumps(canon_cred_wql(json.loads(query_json))))  # indy-sdk requires attr name canonicalization
        chunk = min(cardinality, limit or cardinality, Wallet.DEFAULT_CHUNK)
        if limit:
            cardinality = min(limit, cardinality)
        try:
            while len(infos) != cardinality:
                batch = json.loads(await anoncreds.prover_fetch_credentials(handle, chunk))
                infos.extend(batch)
                if len(batch) < chunk:
                    break
            if len(infos) != cardinality:
                LOGGER.warning('Credential search/limit indicated %s results but fetched %s', cardinality, len(infos))
        finally:
            await anoncreds.prover_close_credentials_search(handle)

        rv_json = json.dumps(infos)
        LOGGER.debug('HolderProver.get_cred_infos_by_q <<< %s', rv_json)
        return rv_json

    async def get_cred_infos_by_filter(self, filt: dict = None) -> str:
        """
        Return cred-info (json list) from wallet by input filter for
        schema identifier and/or credential definition identifier components;
        return info of all credentials for no filter.

        Raise WalletState if the wallet is closed.

        :param filt: indy-sdk filter for credentials; i.e.,

        ::

            {
                "schema_id": string,  # optional
                "schema_issuer_did": string,  # optional
                "schema_name": string,  # optional
                "schema_version": string,  # optional
                "issuer_did": string,  # optional
                "cred_def_id": string  # optional
            }

        :return: credential infos as json list; i.e.,

        ::

            [
                {
                    "referent": string,  # credential identifier in the wallet
                    "attrs": {
                        "attr1" : {"raw": "value1", "encoded": "value1_as_int" },
                        "attr2" : {"raw": "value2", "encoded": "value2_as_int" },
                        ...
                    }
                    "schema_id": string,
                    "cred_def_id": string,
                    "rev_reg_id": Optional<string>,
                    "cred_rev_id": Optional<string>
                },
                ...
            ]

        """

        LOGGER.debug('HolderProver.get_cred_infos_by_filter >>> filt: %s', filt)

        if not self.wallet.handle:
            LOGGER.debug('HolderProver.get_cred_infos_by_filter <!< Wallet %s is closed', self.name)
            raise WalletState('Wallet {} is closed'.format(self.name))

        rv_json = await anoncreds.prover_get_credentials(self.wallet.handle, json.dumps(filt or {}))
        LOGGER.debug('HolderProver.get_cred_infos_by_filter <<< %s', rv_json)
        return rv_json

    async def delete_cred(self, cred_id: str) -> None:
        """
        Delete credential by wallet credential identifier.

        Raise AbsentCred for no such credential. Raise WalletState if the wallet is closed.

        :param cred_id: credential identifier of interest
        """

        LOGGER.debug('HolderProver.delete_cred >>> cred_id: %s', cred_id)

        if not self.wallet.handle:
            LOGGER.debug('HolderProver.delete_cred <!< Wallet %s is closed', self.name)
            raise WalletState('Wallet {} is closed'.format(self.name))

        try:
            await anoncreds.prover_delete_credential(self.wallet.handle, cred_id)
        except IndyError as x_indy:  # no such cred
            if x_indy.error_code == ErrorCode.WalletItemNotFound:
                LOGGER.debug(
                    'HolderProver.delete_cred <!< no cred in wallet %s for cred id %s',
                    self.name,
                    cred_id)
                raise AbsentCred('No cred in wallet for {}'.format(cred_id))
            LOGGER.debug(
                'HolderProver.delete_cred <!< wallet %s, cred id %s: indy error code %s',
                self.name,
                cred_id,
                x_indy.error_code)
            raise

        LOGGER.debug('HolderProver.delete_cred <<<')

    async def get_cred_info_by_id(self, cred_id: str) -> str:
        """
        Return cred-info json from wallet by wallet credential identifier.

        Raise AbsentCred for no such credential. Raise WalletState if the wallet is closed.

        :param cred_id: credential identifier of interest
        :return: json with cred for input credential identifier

        :return: cred-info json; i.e.,

        ::

            {
                "referent": string,  # credential identifier in the wallet
                "attrs": {
                    "attr1" : {"raw": "value1", "encoded": "value1_as_int" },
                    "attr2" : {"raw": "value2", "encoded": "value2_as_int" },
                    ...
                }
                "schema_id": string,
                "cred_def_id": string,
                "rev_reg_id": Optional<string>,
                "cred_rev_id": Optional<string>
            }
        """

        LOGGER.debug('HolderProver.get_cred_info_by_id >>> cred_id: %s', cred_id)

        if not self.wallet.handle:
            LOGGER.debug('HolderProver.get_cred_info_by_id <!< Wallet %s is closed', self.name)
            raise WalletState('Wallet {} is closed'.format(self.name))

        try:
            rv_json = await anoncreds.prover_get_credential(self.wallet.handle, cred_id)
        except IndyError as x_indy:  # no such cred
            if x_indy.error_code == ErrorCode.WalletItemNotFound:
                LOGGER.debug(
                    'HolderProver.get_cred_info_by_id <!< no cred in wallet %s for cred id %s',
                    self.name,
                    cred_id)
                raise AbsentCred('No cred in wallet for {}'.format(cred_id))
            LOGGER.debug(
                'HolderProver.get_cred_info_by_id <!< wallet %s, cred id %s: indy error code %s',
                self.name,
                cred_id,
                x_indy.error_code)
            raise

        LOGGER.debug('HolderProver.get_cred_info_by_id <<< %s', rv_json)
        return rv_json

    async def get_cred_briefs_by_proof_req_q(self, proof_req_json: str, x_queries_json: str = None) -> str:
        """
        A cred-brief aggregates a cred-info and a non-revocation interval. A cred-brief-dict maps
        wallet cred-ids to their corresponding cred-briefs.

        Return json (cred-brief-dict) object mapping wallet credential identifiers to cred-briefs by
        proof request and WQL queries by proof request referent. Return empty dict on no WQL query and
        empty requested predicates specification within proof request. Utility util.proof_req2wql_all()
        builds WQL to retrieve all cred-briefs for (some or all) cred-def-ids in a proof request.

        For each WQL query on an item referent, indy-sdk takes the WQL and the attribute name
        and restrictions (e.g., cred def id, schema id, etc.) from its referent.  Note that
        util.proof_req_attr_referents() maps cred defs and attr names to proof req item referents,
        bridging the gap between attribute names and their corresponding item referents.

        Raise WalletState if the wallet is closed.

        :param proof_req_json: proof request as per Verifier.build_proof_req_json(); e.g.,

        ::

            {
                "nonce": "1532429687",
                "name": "proof_req",
                "version": "0.0",
                "requested_predicates": {},
                "requested_attributes": {
                    "17_name_uuid": {
                        "restrictions": [
                            {
                                "cred_def_id": "LjgpST2rjsoxYegQDRm7EL:3:CL:17:tag"
                            }
                        ],
                        "name": "name"
                    },
                    "17_thing_uuid": {
                        "restrictions": [
                            {
                                "cred_def_id": "LjgpST2rjsoxYegQDRm7EL:3:CL:17:tag"
                            }
                        ],
                        "name": "thing"
                    }
                }
            }

        :param x_queries_json: json list of extra queries to apply to proof request attribute and predicate
            referents; e.g.,

        ::

            {
                "17_thing_uuid": { # require attr presence on name 'thing', cred def id from proof req above
                    "$or": [
                        {
                            "attr::name::value": "J.R. 'Bob' Dobbs"
                        },
                        {
                            "attr::thing::value": "slack"
                        },
                    ]
                },
            }

        :return: json (cred-brief-dict) object mapping wallet cred ids to cred briefs; e.g.,

        ::

            {
                "b42ce5bc-b690-43cd-9493-6fe86ad25e85": {
                    "interval": null,
                    "cred_info": {
                        "schema_id": "LjgpST2rjsoxYegQDRm7EL:2:non-revo:1.0",
                        "rev_reg_id": null,
                        "attrs": {
                            "name": "J.R. \"Bob\" Dobbs",
                            "thing": "slack"
                        },
                        "cred_rev_id": null,
                        "referent": "b42ce5bc-b690-43cd-9493-6fe86ad25e85",
                        "cred_def_id": "LjgpST2rjsoxYegQDRm7EL:3:CL:17:tag"
                    }
                },
                "d773434a-0080-4e3e-a03b-f2033eae7d75": {
                    "interval": null,
                    "cred_info": {
                        "schema_id": "LjgpST2rjsoxYegQDRm7EL:2:non-revo:1.0",
                        "rev_reg_id": null,
                        "attrs": {
                            "name": "Chicken Hawk",
                            "thing": "chicken"
                        },
                        "cred_rev_id": null,
                        "referent": "d773434a-0080-4e3e-a03b-f2033eae7d75",
                        "cred_def_id": "LjgpST2rjsoxYegQDRm7EL:3:CL:17:tag"
                    }
                }
            }

        """

        LOGGER.debug(
            ('HolderProver.get_cred_briefs_by_proof_req_q >>> proof_req_json: %s, x_queries_json: %s'),
            proof_req_json,
            x_queries_json)

        if not self.wallet.handle:
            LOGGER.debug('HolderProver.get_cred_briefs_by_proof_req_q <!< Wallet %s is closed', self.name)
            raise WalletState('Wallet {} is closed'.format(self.name))

        def _pred_filter(brief):
            nonlocal pred_refts
            for attr, preds in pred_refts.get(brief['cred_info']['cred_def_id'], {}).items():
                if any(Predicate.get(p[0]).value.no(brief['cred_info']['attrs'][attr], p[1]) for p in preds.values()):
                    return False
            return True

        rv = {}
        item_refts = set()
        x_queries = json.loads(x_queries_json or '{}')
        for k in x_queries:
            x_queries[k] = canon_cred_wql(x_queries[k])  # indy-sdk requires attr name canonicalization
            item_refts.add(k)

        proof_req = json.loads(proof_req_json)
        item_refts.update(uuid for uuid in proof_req['requested_predicates'])
        if not x_queries:
            item_refts.update(uuid for uuid in proof_req['requested_attributes'])  # get all req attrs if no extra wql
        handle = await anoncreds.prover_search_credentials_for_proof_req(
            self.wallet.handle,
            proof_req_json,
            json.dumps(x_queries) if x_queries else None)
        pred_refts = proof_req_pred_referents(proof_req)

        try:
            for item_referent in item_refts:
                count = Wallet.DEFAULT_CHUNK
                while count == Wallet.DEFAULT_CHUNK:
                    fetched = json.loads(await anoncreds.prover_fetch_credentials_for_proof_req(
                        handle,
                        item_referent,
                        Wallet.DEFAULT_CHUNK))
                    count = len(fetched)
                    for brief in fetched:  # apply predicates from proof req here
                        if brief['cred_info']['referent'] not in rv and _pred_filter(brief):
                            rv[brief['cred_info']['referent']] = brief
        finally:
            await anoncreds.prover_close_credentials_search_for_proof_req(handle)

        rv_json = json.dumps(rv)
        LOGGER.debug('HolderProver.get_cred_briefs_by_proof_req_q <<< %s', rv_json)
        return rv_json


    async def create_proof(self, proof_req: dict, briefs: Union[dict, Sequence[dict]], requested_creds: dict) -> str:
        """
        Create proof as HolderProver.

        Raise:
            * AbsentLinkSecret if link secret not set
            * CredentialFocus on attempt to create proof on no briefs or multiple briefs for a credential definition
            * AbsentTails if missing required tails file
            * | BadRevStateTime if a timestamp for a revocation registry state in the proof request
              | occurs before revocation registry creation
            * IndyError for any other indy-sdk error
            * AbsentInterval if briefs missing non-revocation interval, but cred def supports revocation
            * WalletState if the wallet is closed.

        :param proof_req: proof request as per Verifier.build_proof_req_json()
        :param briefs: cred-brief, sequence thereof, or mapping from wallet cred-id to briefs, to prove
        :param requested_creds: requested credentials data structure; i.e.,

        ::

            {
                'self_attested_attributes': {},
                'requested_attributes': {
                    'attr0_uuid': {
                        'cred_id': string,
                        'timestamp': integer,  # for revocation state
                        'revealed': bool
                    },
                    ...
                },
                'requested_predicates': {
                    'predicate0_uuid': {
                        'cred_id': string,
                        'timestamp': integer  # for revocation state
                    }
                }
            }

        :return: proof json
        """

        LOGGER.debug(
            'HolderProver.create_proof >>> proof_req: %s, briefs: %s, requested_creds: %s',
            proof_req,
            briefs,
            requested_creds)

        if not self.wallet.handle:
            LOGGER.debug('HolderProver.create_proof <!< Wallet %s is closed', self.name)
            raise WalletState('Wallet {} is closed'.format(self.name))

        label = await self._assert_link_secret('create_proof')

        cd_ids = set()
        x_cd_ids = set()
        for brief in iter_briefs(briefs):
            cd_id = brief['cred_info']['cred_def_id']
            if cd_id in cd_ids and cd_id not in x_cd_ids:
                x_cd_ids.add(cd_id)
            cd_ids.add(cd_id)
            if x_cd_ids:
                LOGGER.debug('HolderProver.create_proof <!< briefs specification out of focus (non-uniqueness)')
                raise CredentialFocus('Briefs list repeats cred defs: {}'.format(x_cd_ids))

        s_id2schema = {}  # schema identifier to schema
        cd_id2cred_def = {}  # credential definition identifier to credential definition
        rr_id2timestamp = {}  # revocation registry of interest to timestamp of interest (or None)
        rr_id2cr_id = {}  # revocation registry of interest to credential revocation identifier
        for brief in iter_briefs(briefs):
            interval = brief.get('interval', None)
            cred_info = brief['cred_info']
            s_id = cred_info['schema_id']
            if not ok_schema_id(s_id):
                LOGGER.debug('HolderProver.create_proof <!< Bad schema id %s', s_id)
                raise BadIdentifier('Bad schema id {}'.format(s_id))

            if s_id not in s_id2schema:
                schema = json.loads(await self.get_schema(s_id))  # add to cache en passant
                if not schema:
                    LOGGER.debug(
                        'HolderProver.create_proof <!< absent schema %s, proof req may be for another ledger',
                        s_id)
                    raise AbsentSchema('Absent schema {}, proof req may be for another ledger'.format(s_id))
                s_id2schema[s_id] = schema

            cd_id = cred_info['cred_def_id']
            if not ok_cred_def_id(cd_id):
                LOGGER.debug('HolderProver.create_proof <!< Bad cred def id %s', cd_id)
                raise BadIdentifier('Bad cred def id {}'.format(cd_id))

            if cd_id not in cd_id2cred_def:
                cred_def = json.loads(await self.get_cred_def(cd_id))  # add to cache en passant
                cd_id2cred_def[cd_id] = cred_def

            rr_id = cred_info['rev_reg_id']
            if rr_id:
                if not ok_rev_reg_id(rr_id):
                    LOGGER.debug('HolderProver.create_proof <!< Bad rev reg id %s', rr_id)
                    raise BadIdentifier('Bad rev reg id {}'.format(rr_id))

                await self._sync_revoc_for_proof(rr_id)  # link tails file to its rr_id if it's new
                if interval:
                    if rr_id not in rr_id2timestamp:
                        if interval['to'] > int(time()):
                            LOGGER.debug(
                                'HolderProver.create_proof <!< interval to %s for rev reg %s is in the future',
                                interval['to'],
                                rr_id)
                            raise BadRevStateTime(
                                'Revocation registry {} timestamp {} is in the future'.format(rr_id, interval['to']))
                        rr_id2timestamp[rr_id] = interval['to']
                elif 'revocation' in cd_id2cred_def[cd_id]['value']:
                    LOGGER.debug(
                        'HolderProver.create_proof <!< brief on cred def id %s missing non-revocation interval',
                        cd_id)
                    raise AbsentInterval('Brief on cred def id {} missing non-revocation interval'.format(cd_id))
                if rr_id in rr_id2cr_id:
                    continue
                rr_id2cr_id[rr_id] = cred_info['cred_rev_id']

        rr_id2rev_state = {}  # revocation registry identifier to its state
        with REVO_CACHE.lock:
            for rr_id in rr_id2timestamp:
                revo_cache_entry = REVO_CACHE.get(rr_id, None)
                tails = revo_cache_entry.tails if revo_cache_entry else None
                if tails is None:  # missing tails file
                    LOGGER.debug('HolderProver.create_proof <!< missing tails file for rev reg id %s', rr_id)
                    raise AbsentTails('Missing tails file for rev reg id {}'.format(rr_id))
                rr_def_json = await self.get_rev_reg_def(rr_id)
                (rr_delta_json, ledger_timestamp) = await revo_cache_entry.get_delta_json(
                    self._build_rr_delta_json,
                    rr_id2timestamp[rr_id],
                    rr_id2timestamp[rr_id])
                rr_state_json = await anoncreds.create_revocation_state(
                    tails.reader_handle,
                    rr_def_json,
                    rr_delta_json,
                    ledger_timestamp,
                    rr_id2cr_id[rr_id])
                rr_id2rev_state[rr_id] = {
                    rr_id2timestamp[rr_id]: json.loads(rr_state_json)
                }

        rv = await anoncreds.prover_create_proof(
            self.wallet.handle,
            json.dumps(proof_req),
            json.dumps(requested_creds),
            label,
            json.dumps(s_id2schema),
            json.dumps(cd_id2cred_def),
            json.dumps(rr_id2rev_state))
        LOGGER.debug('HolderProver.create_proof <<< %s', rv)
        return rv

    async def reset_wallet(self, seed: str = None) -> Wallet:
        """
        Close and delete HolderProver wallet, then create and open a replacement on prior link secret.
        Note that this operation effectively destroys private keys for credential definitions. Its
        intended use is primarily for testing and demonstration.

        Raise AbsentLinkSecret if link secret not set. Raise WalletState if the wallet is closed.

        :param seed: seed to use for new wallet (default random)
        :return: replacement wallet
        """

        LOGGER.debug('HolderProver.reset_wallet >>>')

        self.wallet = await WalletManager().reset(self.wallet, seed)

        rv = self.wallet
        LOGGER.debug('HolderProver.reset_wallet <<< %s', rv)
        return rv
